#!/usr/bin/env python3
"""
Reward system unit testing
"""

import pytest
import numpy as np
from lib.reward import RewardCalculator, RewardSignal


class TestRewardSignal:
    """testRewardSignalkind"""

    def test_collect_normalize(self):
        """Test signal collection and normalization"""
        signal = RewardSignal(
            name="test",
            weight=0.5,
            collector=lambda ctx: ctx.get("value", 0),
            min_value=0.0,
            max_value=10.0
        )

        # Test the middle value
        result = signal.collect({"value": 5.0})
        assert result == 0.5

        # Test maximum
        result = signal.collect({"value": 10.0})
        assert result == 1.0

        # Test minimum
        result = signal.collect({"value": 0.0})
        assert result == 0.0

        # test out of range（Truncate）
        result = signal.collect({"value": 15.0})
        assert result == 1.0

        result = signal.collect({"value": -5.0})
        assert result == 0.0

    def test_history_tracking(self):
        """Test history"""
        signal = RewardSignal(
            name="test",
            weight=0.5,
            collector=lambda ctx: ctx.get("value", 0),
            min_value=0.0,
            max_value=10.0
        )

        signal.collect({"value": 3.0})
        signal.collect({"value": 7.0})

        assert len(signal.history) == 2
        assert signal.history[0] == 0.3
        assert signal.history[1] == 0.7

    def test_update_weight(self):
        """Test weight update"""
        signal = RewardSignal(
            name="test",
            weight=0.5,
            collector=lambda ctx: 0,
            min_value=0.0,
            max_value=1.0
        )

        signal.update_weight(0.1)
        assert signal.weight == 0.6

        signal.update_weight(-0.3)
        assert signal.weight == 0.3

        # test boundaries
        signal.update_weight(1.0)
        assert signal.weight == 1.0

        signal.update_weight(-2.0)
        assert signal.weight == 0.0


class TestRewardCalculator:
    """testRewardCalculatorkind"""

    def test_initialization(self):
        """Test initialization"""
        calc = RewardCalculator()

        # Check that all signals are initialized
        expected_signals = [
            "test_coverage", "code_quality", "bug_count", "task_time",
            "acceptance_rate", "adoption_rate", "rewrite_rate",
            "user_rating", "feedback_count",
            "agent_preference", "workflow_preference"
        ]

        for signal_name in expected_signals:
            assert signal_name in calc.signals

    def test_calculate_reward(self):
        """Test reward calculation"""
        calc = RewardCalculator()

        context = {
            "task_type": "T2",
            "task_result": {
                "agent": "claude",
                "workflow": "tdd",
                "duration": 300,
                "test_files_created": True,
                "committed": True
            },
            "test_result": {
                "coverage": 80.0,
                "passed": 10,
                "failed": 0
            },
            "user_feedback": {
                "accepted": True,
                "rating": 5
            },
            "metrics": {
                "complexity": 3,
                "duplication": 0.05,
                "lint_score": 0.9
            }
        }

        reward = calc.calculate_reward(context)

        # The reward should be in [0, 1] within range
        assert 0.0 <= reward <= 1.0

        # High-quality tasks should have higher rewards
        assert reward > 0.5

    def test_weight_adjustment(self):
        """Test weight adjustment after negative feedback"""
        calc = RewardCalculator()

        # Record initial weight
        initial_weights = {
            name: signal.weight
            for name, signal in calc.signals.items()
        }

        # Record negative feedback
        calc.record_feedback("rating", 2)

        # Check that the weights have been adjusted
        # Notice：due to normalization，We examine relative changes rather than absolute increases or decreases
        agent_auto_weights = ["agent_preference", "workflow_preference"]
        feedback_quality_weights = ["user_rating", "feedback_count", "code_quality", "test_coverage"]

        # Calculate average change
        agent_auto_change = np.mean([
            (calc.signals[name].weight - initial_weights[name]) / initial_weights[name]
            for name in agent_auto_weights
        ])
        feedback_quality_change = np.mean([
            (calc.signals[name].weight - initial_weights[name]) / initial_weights[name]
            for name in feedback_quality_weights
        ])

        # The average change in automation weights should be less than the feedback/quality weight
        assert agent_auto_change < feedback_quality_change

    def test_weight_normalization(self):
        """Test weight normalization"""
        calc = RewardCalculator(learning_phase="early")
        total_weight = sum(signal.weight for signal in calc.signals.values())
        assert total_weight == pytest.approx(1.0, rel=1e-3)

    def test_negative_signals_reduce_reward(self):
        """Testing the impact of negative signals on rewards"""
        calc = RewardCalculator()

        base_context = {
            "task_result": {"duration": 100, "errors": 0},
            "test_result": {"coverage": 90.0, "failed": 0},
            "user_feedback": {"accepted": True, "rating": 5},
            "metrics": {"complexity": 2, "duplication": 0.0, "lint_score": 1.0}
        }

        high_bug_context = {
            "task_result": {"duration": 1000, "errors": 5},
            "test_result": {"coverage": 90.0, "failed": 5},
            "user_feedback": {"accepted": True, "rating": 5},
            "metrics": {"complexity": 2, "duplication": 0.0, "lint_score": 1.0}
        }

        reward_base = calc.calculate_reward(base_context)
        reward_high_bug = calc.calculate_reward(high_bug_context)

        assert reward_high_bug < reward_base

    def test_learning_phase_transition(self):
        """Test learning phase transition"""
        calc = RewardCalculator(learning_phase="early")

        # The initial stage should beearly
        assert calc.learning_phase == "early"

        # simulation20tasks
        for i in range(20):
            context = {
                "task_result": {},
                "test_result": {},
                "user_feedback": {},
                "metrics": {}
            }
            calc.calculate_reward(context)

        # should be converted tomaturestage
        assert calc.learning_phase == "mature"

    def test_reward_stats(self):
        """Test reward statistics"""
        calc = RewardCalculator()

        # Generate some bonus data
        for i in range(10):
            context = {
                "task_result": {"duration": 300},
                "test_result": {"coverage": 50 + i * 5},
                "user_feedback": {"accepted": True},
                "metrics": {}
            }
            calc.calculate_reward(context)

        stats = calc.get_reward_stats()

        assert "mean" in stats
        assert "std" in stats
        assert "min" in stats
        assert "max" in stats
        assert stats["count"] == 10

    def test_signal_stats(self):
        """Test individual signal statistics"""
        calc = RewardCalculator()

        # generate some data
        for i in range(5):
            context = {
                "task_result": {},
                "test_result": {"coverage": 50 + i * 10},
                "user_feedback": {},
                "metrics": {}
            }
            calc.calculate_reward(context)

        stats = calc.get_signal_stats("test_coverage")

        assert stats["count"] == 5
        assert "mean" in stats
        assert "current_weight" in stats

    def test_save_load_state(self):
        """Test state saving and loading"""
        import tempfile
        import os

        calc1 = RewardCalculator()

        # generate some data
        for i in range(3):
            context = {
                "task_result": {"duration": 300},
                "test_result": {"coverage": 50 + i * 10},
                "user_feedback": {"rating": 4},
                "metrics": {}
            }
            calc1.calculate_reward(context)

        # save state
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.json') as f:
            temp_path = f.name

        try:
            calc1.save_state(temp_path)

            # Create new calculator and load status
            calc2 = RewardCalculator()
            calc2.load_state(temp_path)

            # Check status restored
            assert calc2.task_count == calc1.task_count
            assert len(calc2.reward_history) == len(calc1.reward_history)
            assert calc2.learning_phase == calc1.learning_phase

        finally:
            os.unlink(temp_path)


class TestRewardCollectors:
    """Test individual data collection functions"""

    def test_collect_test_coverage(self):
        """Test test coverage collection"""
        calc = RewardCalculator()

        # There are test results
        context = {
            "test_result": {"coverage": 85.0}
        }
        assert calc._collect_test_coverage(context) == 85.0

        # No test results，But there is a test file created
        context = {
            "task_result": {"test_files_created": True}
        }
        assert calc._collect_test_coverage(context) == 50.0

    def test_collect_code_quality(self):
        """Test code quality collection"""
        calc = RewardCalculator()

        context = {
            "metrics": {
                "complexity": 3,
                "duplication": 0.1,
                "lint_score": 0.8
            }
        }

        quality = calc._collect_code_quality(context)

        # should be in [0, 10] within range
        assert 0.0 <= quality <= 10.0

        # low complexity、Low repetition、highlintShould get high quality points
        assert quality > 5.0

    def test_collect_bug_count(self):
        """testBugquantity collection"""
        calc = RewardCalculator()

        context = {
            "test_result": {"failed": 2},
            "task_result": {"errors": 1}
        }

        bugs = calc._collect_bug_count(context)
        assert bugs == 3

    def test_collect_task_time(self):
        """Test task time collection"""
        calc = RewardCalculator()

        context = {
            "task_result": {"duration": 600}
        }

        time = calc._collect_task_time(context)
        assert time == 600

    def test_collect_acceptance_rate(self):
        """Test acceptance rate collection"""
        calc = RewardCalculator()

        # accept
        context = {
            "user_feedback": {"accepted": True}
        }
        assert calc._collect_acceptance_rate(context) == 1.0

        # reject
        context = {
            "user_feedback": {"accepted": False}
        }
        assert calc._collect_acceptance_rate(context) == 0.0

        # There are modifications
        context = {
            "user_feedback": {"revisions": 3}
        }
        rate = calc._collect_acceptance_rate(context)
        assert rate < 1.0

    def test_collect_user_rating(self):
        """Test user rating collection"""
        calc = RewardCalculator()

        # Explicit scoring
        context = {
            "user_feedback": {"rating": 5}
        }
        assert calc._collect_user_rating(context) == 5

        # fromacceptanceinfer
        context = {
            "user_feedback": {"accepted": True}
        }
        assert calc._collect_user_rating(context) == 4.0

    def test_collect_agent_preference(self):
        """testAgentPreference collection"""
        calc = RewardCalculator()

        # T2Task usageClaude
        context = {
            "task_type": "T2",
            "task_result": {"agent": "claude"}
        }
        preference = calc._collect_agent_preference(context)
        assert preference > 0.5

        # T3Task usageCodex
        context = {
            "task_type": "T3",
            "task_result": {"agent": "codex"}
        }
        preference = calc._collect_agent_preference(context)
        assert preference > 0.5

    def test_collect_workflow_preference(self):
        """Test workflow preference collection"""
        calc = RewardCalculator()

        # TDDWorkflow
        context = {
            "task_result": {"workflow": "tdd"}
        }
        preference = calc._collect_workflow_preference(context)
        assert preference > 0.8

        # Standard workflow
        context = {
            "task_result": {"workflow": "standard"}
        }
        preference = calc._collect_workflow_preference(context)
        assert preference < 0.8


class StubHistoryProvider:
    """History provider for testing"""

    def __init__(self, agent_rate: float | None = None, workflow_rate: float | None = None):
        self.agent_rate = agent_rate
        self.workflow_rate = workflow_rate

    def get_agent_success_rate(self, task_type: str, agent: str):
        return self.agent_rate

    def get_workflow_success_rate(self, task_type: str, workflow: str):
        return self.workflow_rate


class TestRewardHistoryProvider:
    """Testing history preference provider access"""

    def test_history_provider_overrides_agent_preference(self):
        calc = RewardCalculator(history_provider=StubHistoryProvider(agent_rate=0.92))
        context = {"task_type": "T3", "task_result": {"agent": "codex"}}
        assert calc._collect_agent_preference(context) == pytest.approx(0.92)

    def test_history_provider_overrides_workflow_preference(self):
        calc = RewardCalculator(history_provider=StubHistoryProvider(workflow_rate=0.88))
        context = {"task_type": "T2", "task_result": {"workflow": "tdd"}}
        assert calc._collect_workflow_preference(context) == pytest.approx(0.88)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



#!/usr/bin/env python3
"""Risk assessment helpers for confirmation decisions."""

from __future__ import annotations

import re
from enum import Enum
from typing import Any, Mapping, Sequence


class RiskLevel(Enum):
    """Risk categories for task execution."""

    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"
    CRITICAL = "critical"


def _safe_text(task_context: Mapping[str, Any], key: str) -> str:
    value = task_context.get(key, "")
    if isinstance(value, str):
        return value
    return str(value)


def _safe_files(task_context: Mapping[str, Any]) -> list[str]:
    raw_files = task_context.get("files", [])
    if not isinstance(raw_files, Sequence) or isinstance(raw_files, (str, bytes)):
        return []
    return [str(item) for item in raw_files]


def _normalize_text(value: str) -> str:
    return " ".join(value.lower().strip().split())


def _max_risk(current: RiskLevel, new: RiskLevel) -> RiskLevel:
    order = {
        RiskLevel.LOW: 0,
        RiskLevel.MEDIUM: 1,
        RiskLevel.HIGH: 2,
        RiskLevel.CRITICAL: 3,
    }
    return new if order[new] > order[current] else current


class RiskAssessor:
    """Classify task contexts into coarse execution risk buckets."""

    def __init__(self) -> None:
        self.high_risk_keywords = [
            "rm",
            "remove",
            "delete",
            "unlink",
            "rmdir",
            "format",
            "fdisk",
            "mkfs",
            "chmod",
            "chown",
            "iptables",
            "docker",
            "podman",
            "kubectl",
            "git push --force",
            "git reset --hard",
        ]
        self.medium_risk_keywords = [
            "database",
            "migrate",
            "schema",
            "api",
            "fetch",
            "request",
            "install",
            "uninstall",
            "update",
        ]
        self.low_risk_keywords = [
            "test",
            "tests",
            "format",
            "lint",
            "check",
            "log",
            "status",
            "info",
            "list",
            "validate",
            "verify",
            "docs",
            "read",
            "view",
            "show",
            "display",
            "inspect",
            "describe",
            "explain",
            "analyze",
            "measure",
            "benchmark",
            "profile",
            "build",
            "compile",
            "bundle",
        ]

    def assess(self, task_context: Mapping[str, Any]) -> RiskLevel:
        """Return only the risk bucket for one task."""
        risk_level, _ = self.assess_details(task_context)
        return risk_level

    def assess_details(self, task_context: Mapping[str, Any]) -> tuple[RiskLevel, list[str]]:
        """Assess risk and return the heuristic basis for that assessment."""
        task_desc = _normalize_text(_safe_text(task_context, "task_description"))
        command = _normalize_text(_safe_text(task_context, "command"))
        files = _safe_files(task_context)
        combined_text = f"{task_desc} {command}".strip()

        risk_level = RiskLevel.MEDIUM
        basis: list[str] = []

        context_risk, context_basis = self._assess_command_context(command)
        if context_basis:
            basis.extend(context_basis)
            risk_level = _max_risk(risk_level, context_risk)

        for keyword in self.high_risk_keywords:
            if keyword == "format":
                if any(ctx in combined_text for ctx in ["mkfs", "fdisk", "format disk", "format /dev"]):
                    basis.append("HIGH risk keyword: format disk")
                    risk_level = _max_risk(risk_level, RiskLevel.CRITICAL)
                continue
            if " " in keyword:
                if keyword in combined_text:
                    basis.append(f"HIGH risk keyword: {keyword}")
                    risk_level = _max_risk(risk_level, RiskLevel.CRITICAL)
                continue
            pattern = r"\b" + re.escape(keyword) + r"\b"
            if re.search(pattern, combined_text):
                basis.append(f"HIGH risk keyword: {keyword}")
                risk_level = _max_risk(risk_level, RiskLevel.CRITICAL)

        for keyword in self.medium_risk_keywords:
            pattern = r"\b" + re.escape(keyword) + r"\b"
            if re.search(pattern, combined_text):
                basis.append(f"MEDIUM risk keyword: {keyword}")
                risk_level = _max_risk(risk_level, RiskLevel.MEDIUM)

        for keyword in self.low_risk_keywords:
            pattern = r"\b" + re.escape(keyword) + r"\b"
            if re.search(pattern, combined_text):
                basis.append(f"LOW risk keyword: {keyword}")
                risk_level = RiskLevel.LOW if risk_level == RiskLevel.MEDIUM else risk_level

        file_risk, file_basis = self._assess_file_operation_risk_details(files)
        if file_basis:
            basis.extend(file_basis)
            risk_level = _max_risk(risk_level, file_risk)

        if not basis:
            basis.append("No strong heuristic match")

        return risk_level, basis

    def _assess_command_context(self, command: str) -> tuple[RiskLevel, list[str]]:
        if "git" in command:
            if any(item in command for item in ["git status", "git log", "git diff", "git show"]):
                return RiskLevel.LOW, ["LOW risk command context: git read-only"]
            if "git push" in command and "--force" not in command:
                return RiskLevel.MEDIUM, ["MEDIUM risk command context: git push"]
            if any(item in command for item in ["git commit", "git add"]):
                return RiskLevel.LOW, ["LOW risk command context: git local write"]

        if any(item in command for item in ["npm run", "pnpm run", "yarn"]):
            if any(item in command for item in ["test", "lint", "format", "check", "build"]):
                return RiskLevel.LOW, ["LOW risk command context: package manager verification command"]

        if "python" in command or "pytest" in command:
            if any(item in command for item in ["-m pytest", "pytest", "mypy", "black", "ruff"]):
                return RiskLevel.LOW, ["LOW risk command context: Python verification command"]

        return RiskLevel.MEDIUM, []

    def _assess_file_operation_risk_details(self, files: Sequence[str]) -> tuple[RiskLevel, list[str]]:
        if not files:
            return RiskLevel.MEDIUM, []

        basis: list[str] = []
        risk_level = RiskLevel.MEDIUM
        for file_path in files:
            normalized = _normalize_text(str(file_path))
            if any(item in normalized for item in ["home", "etc", "system", "root"]):
                basis.append(f"HIGH risk file path: {file_path}")
                risk_level = _max_risk(risk_level, RiskLevel.HIGH)
            if any(item in normalized for item in [".env", "config.py", "config.json", "secrets"]):
                basis.append(f"HIGH risk file path: {file_path}")
                risk_level = _max_risk(risk_level, RiskLevel.HIGH)
            if any(item in normalized for item in ["package.json", "requirements.txt", "pom.xml", "build.gradle"]):
                basis.append(f"MEDIUM risk file path: {file_path}")
                risk_level = _max_risk(risk_level, RiskLevel.MEDIUM)
            if any(item in normalized for item in ["delete", "remove", "unlink"]):
                basis.append(f"CRITICAL risk file path: {file_path}")
                risk_level = _max_risk(risk_level, RiskLevel.CRITICAL)
        return risk_level, basis



#!/usr/bin/env python3
"""Regression tests for ConfirmationAPI bootstrap behavior."""

from __future__ import annotations

from pathlib import Path

from lib.api import ConfirmationAPI
from lib.cli import ClawPolicyCLI
from lib.policy_store import PolicyStore


def test_confirmation_api_bootstraps_policy_store_when_missing(tmp_path: Path) -> None:
    memory_dir = tmp_path / ".clawpolicy"

    api = ConfirmationAPI(memory_dir=memory_dir)

    assert api.policy_store is not None
    assert not hasattr(api, "gep_store")
    assert (memory_dir / "policy").exists()
    assert (memory_dir / "policy" / "rules.json").exists()
    assert (memory_dir / "policy" / "playbooks.json").exists()
    assert (memory_dir / "policy" / "policy_events.jsonl").exists()


def test_confirmation_api_returns_structured_decision_details(tmp_path: Path) -> None:
    memory_dir = tmp_path / ".clawpolicy"
    api = ConfirmationAPI(memory_dir=memory_dir)

    task = {
        "task_type": "T2",
        "task_description": "Run tests",
        "command": "python -m pytest tests/",
    }

    auto_execute, reason, details = api.should_auto_execute(task)

    assert auto_execute is True
    assert reason
    assert details["decision_id"]
    assert details["decision"]["final_decision"] == "auto_execute"
    assert "heuristic_basis" in details["decision"]
    assert "matched_rules" in details["decision"]
    assert details["relevant_rules_count"] == details["decision"]["confidence"]["matched_rule_count"]


def test_confirmation_api_persists_decision_history_with_outcome(tmp_path: Path) -> None:
    memory_dir = tmp_path / ".clawpolicy"
    api = ConfirmationAPI(memory_dir=memory_dir)

    task = {
        "task_type": "T2",
        "task_description": "Run tests",
        "command": "python -m pytest tests/",
    }

    _, _, details = api.should_auto_execute(task)
    api.record_execution_result(
        task=task,
        success=True,
        auto_executed=True,
        decision_id=details["decision_id"],
        execution_result="success",
    )

    history = api.get_recent_decisions(limit=5)
    assert history["decisions"]
    latest = history["decisions"][0]
    assert latest["decision_id"] == details["decision_id"]
    assert latest["execution_result"] == "success"
    assert latest["final_decision"] == "auto_execute"
    assert latest["explanation"]["final_decision"] == "auto_execute"


def test_confirmation_api_records_explicit_override_as_policy_candidate(tmp_path: Path) -> None:
    memory_dir = tmp_path / ".clawpolicy"
    api = ConfirmationAPI(memory_dir=memory_dir)

    task = {
        "task_type": "T2",
        "task_description": "Run tests",
        "command": "python -m pytest tests/",
    }

    _, _, details = api.should_auto_execute(task)
    api.record_execution_result(
        task=task,
        success=True,
        auto_executed=False,
        decision_id=details["decision_id"],
        execution_result="success",
        user_override="confirmed_after_prompt",
    )

    rules = PolicyStore(memory_dir / "policy").load_rules()
    assert any(rule.status == "candidate" for rule in rules.values())


def test_decision_history_cli_shows_recent_trace(tmp_path: Path, capsys) -> None:
    memory_dir = tmp_path / ".clawpolicy"
    api = ConfirmationAPI(memory_dir=memory_dir)
    task = {
        "task_type": "T2",
        "task_description": "Run tests",
        "command": "python -m pytest tests/",
    }

    _, _, details = api.should_auto_execute(task)
    api.record_execution_result(
        task=task,
        success=True,
        auto_executed=True,
        decision_id=details["decision_id"],
        execution_result="success",
    )

    cli = ClawPolicyCLI()
    import os

    original_cwd = os.getcwd()
    try:
        os.chdir(tmp_path)
        cli.decision_history(limit=5)
    finally:
        os.chdir(original_cwd)

    output = capsys.readouterr().out
    assert details["decision_id"] in output
    assert "auto_execute" in output
    assert "Run tests" in output


def test_status_cli_uses_policy_terminology(tmp_path: Path, capsys) -> None:
    cli = ClawPolicyCLI()

    import os

    original_cwd = os.getcwd()
    try:
        os.chdir(tmp_path)
        cli.init(force=True)
        cli.status()
    finally:
        os.chdir(original_cwd)

    output = capsys.readouterr().out
    assert "Policy" in output
    assert "Rules" in output
    assert "Playbooks" in output


def test_policy_status_and_suspended_cli_show_recent_lifecycle(tmp_path: Path, capsys) -> None:
    memory_dir = tmp_path / ".clawpolicy"
    api = ConfirmationAPI(memory_dir=memory_dir)
    store = PolicyStore(memory_dir / "policy")
    project_scope = str(tmp_path.resolve())

    confirmed_rule = store.get_rule("confirmed_cli_rule")
    if confirmed_rule is None:
        from lib.policy_models import Rule

        confirmed_rule = Rule(
            id="confirmed_cli_rule",
            summary="Project pytest runs may auto execute",
            category="harden",
            trigger=["task_type:T2", "command:python -m pytest tests/"],
            status="confirmed",
            scope="project",
            scope_key=project_scope,
            evidence_count=4,
            source_type="explicit_preference",
            last_seen_at="2026-03-07T12:00:00+00:00",
            policy_decision="auto_execute",
            validation=["assert auto execute for pytest"],
        )
        confirmed_rule.calculate_asset_id()
        store.save_rule(confirmed_rule)

    task = {
        "task_type": "T2",
        "task_description": "Run tests",
        "command": "python -m pytest tests/",
        "project_path": project_scope,
    }

    for _ in range(2):
        _, _, details = api.should_auto_execute(task)
        api.record_execution_result(
            task=task,
            success=False,
            auto_executed=True,
            decision_id=details["decision_id"],
            execution_result="failure",
        )

    cli = ClawPolicyCLI()
    import os

    original_cwd = os.getcwd()
    try:
        os.chdir(tmp_path)
        cli.policy_status()
        cli.policy_suspended(limit=5)
    finally:
        os.chdir(original_cwd)

    output = capsys.readouterr().out
    assert "confirmed" in output.lower()
    assert "suspended" in output.lower()
    assert "confirmed_cli_rule" in output



#!/usr/bin/env python3
"""Persistence layer for canonical policy assets."""

from __future__ import annotations

import json
from pathlib import Path
from typing import Any

from .policy_models import Playbook, PolicyEvent, Rule


class PolicyStore:
    """Storage manager for rules, playbooks, and policy events."""

    def __init__(self, base_dir: Path):
        self.base_dir = Path(base_dir)
        self.base_dir.mkdir(parents=True, exist_ok=True)

        self.rules_file = self.base_dir / "rules.json"
        self.playbooks_file = self.base_dir / "playbooks.json"
        self.policy_events_file = self.base_dir / "policy_events.jsonl"

    @classmethod
    def bootstrap(cls, memory_dir: Path, *, ensure_files: bool = False) -> "PolicyStore":
        """Open canonical policy storage for one memory directory."""
        store = cls(Path(memory_dir) / "policy")
        if ensure_files:
            if not store.rules_file.exists():
                store.save_rules({})
            if not store.playbooks_file.exists():
                store.save_playbooks({})
            if not store.policy_events_file.exists():
                store.policy_events_file.touch()
        return store

    @staticmethod
    def _read_json_mapping(path: Path) -> dict[str, Any] | None:
        if not path.exists():
            return None

        try:
            return json.loads(path.read_text(encoding="utf-8"))
        except Exception as exc:
            print(f"⚠️  Failed to load {path.name}: {exc}")
            return None

    @staticmethod
    def _atomic_write_json(path: Path, data: dict[str, Any]) -> None:
        temp_file = path.with_suffix(".tmp")
        temp_file.write_text(json.dumps(data, indent=2, ensure_ascii=False), encoding="utf-8")
        temp_file.replace(path)

    def load_rules(self) -> dict[str, Rule]:
        """Load all rules from canonical storage."""
        data = self._read_json_mapping(self.rules_file) or {}
        return {rule_id: Rule.from_dict(rule_data) for rule_id, rule_data in data.items()}

    def save_rules(self, rules: dict[str, Rule]) -> None:
        """Persist canonical rules to disk."""
        data = {rule_id: rule.to_dict() for rule_id, rule in rules.items()}
        self._atomic_write_json(self.rules_file, data)

    def load_playbooks(self) -> dict[str, Playbook]:
        """Load all playbooks from canonical storage."""
        data = self._read_json_mapping(self.playbooks_file) or {}
        return {
            playbook_id: Playbook.from_dict(playbook_data)
            for playbook_id, playbook_data in data.items()
        }

    def save_playbooks(self, playbooks: dict[str, Playbook]) -> None:
        """Persist canonical playbooks to disk."""
        data = {
            playbook_id: playbook.to_dict()
            for playbook_id, playbook in playbooks.items()
        }
        self._atomic_write_json(self.playbooks_file, data)

    def _read_event_lines(self) -> list[str]:
        if not self.policy_events_file.exists():
            return []
        return self.policy_events_file.read_text(encoding="utf-8").splitlines()

    def append_event(self, event: PolicyEvent) -> None:
        """Append one policy event to canonical event storage."""
        if not self.policy_events_file.exists():
            self.policy_events_file.touch()
        with open(self.policy_events_file, "a", encoding="utf-8") as handle:
            handle.write(event.to_jsonl() + "\n")

    def get_events(self, limit: int = 100) -> list[PolicyEvent]:
        """Return latest events in reverse chronological order."""
        events: list[PolicyEvent] = []
        try:
            lines = self._read_event_lines()
            for line in reversed(lines[-limit:]):
                payload = line.strip()
                if not payload:
                    continue
                try:
                    events.append(PolicyEvent.from_jsonl(payload))
                except Exception:
                    continue
            return events
        except Exception as exc:
            print(f"⚠️  Failed to read policy events: {exc}")
            return []

    def get_events_by_type(self, event_types: set[str], limit: int = 100) -> list[PolicyEvent]:
        """Return latest events matching a set of event types."""
        return [event for event in self.get_events(limit=limit) if event.event_type in event_types]

    def get_policy_events(self, limit: int = 100) -> list[PolicyEvent]:
        """Return latest policy events."""
        return self.get_events(limit=limit)

    def get_decision_events(self, limit: int = 100) -> list[PolicyEvent]:
        """Return latest decision evaluation/outcome events."""
        return self.get_events_by_type({"decision_evaluated", "decision_outcome"}, limit=limit)

    def get_recent_lifecycle_events(self, limit: int = 100) -> list[PolicyEvent]:
        """Return recent lifecycle transition events."""
        return self.get_events_by_type(
            {"rule_promoted", "rule_suspended", "rule_reactivated", "rule_archived"},
            limit=limit,
        )

    def get_rules_by_status(self, status: str) -> list[Rule]:
        """Return rules filtered by lifecycle status."""
        return [rule for rule in self.load_rules().values() if rule.status == status]

    def get_risky_confirmed_rules(self) -> list[Rule]:
        """Return confirmed high-risk rules that currently auto execute."""
        return [
            rule
            for rule in self.load_rules().values()
            if rule.status == "confirmed"
            and rule.policy_decision == "auto_execute"
            and rule.risk_level in {"high", "critical"}
        ]

    def get_policy_status_snapshot(self) -> dict[str, Any]:
        """Return a lightweight operational snapshot for policy lifecycle monitoring."""
        rules = list(self.load_rules().values())
        lifecycle_events = self.get_recent_lifecycle_events(limit=100)
        status_counts = {
            status: sum(1 for rule in rules if rule.status == status)
            for status in ["hint", "candidate", "confirmed", "suspended", "archived"]
        }

        special_scopes = [
            {
                "id": rule.id,
                "scope": rule.scope,
                "scope_key": rule.scope_key,
                "status": rule.status,
                "policy_decision": rule.policy_decision,
            }
            for rule in rules
            if rule.scope in {"domain", "project"} and rule.status in {"confirmed", "suspended"}
        ]

        return {
            "status_counts": status_counts,
            "recent_promotions": [
                event.payload for event in lifecycle_events if event.event_type == "rule_promoted"
            ],
            "recent_suspensions": [
                event.payload for event in lifecycle_events if event.event_type == "rule_suspended"
            ],
            "risky_confirmed_rules": [
                {
                    "id": rule.id,
                    "summary": rule.summary,
                    "risk_level": rule.risk_level,
                    "scope": rule.scope,
                    "scope_key": rule.scope_key,
                }
                for rule in self.get_risky_confirmed_rules()
            ],
            "special_scopes": special_scopes,
        }

    def get_rule(self, rule_id: str) -> Rule | None:
        """Return one rule by id."""
        return self.load_rules().get(rule_id)

    def save_rule(self, rule: Rule) -> None:
        """Insert or update one rule."""
        rules = self.load_rules()
        rules[rule.id] = rule
        self.save_rules(rules)

    def get_playbook(self, playbook_id: str) -> Playbook | None:
        """Return one playbook by id."""
        return self.load_playbooks().get(playbook_id)

    def save_playbook(self, playbook: Playbook) -> None:
        """Insert or update one playbook."""
        playbooks = self.load_playbooks()
        playbooks[playbook.id] = playbook
        self.save_playbooks(playbooks)

    def delete_rule(self, rule_id: str) -> bool:
        """Delete one rule if it exists."""
        rules = self.load_rules()
        if rule_id not in rules:
            return False
        del rules[rule_id]
        self.save_rules(rules)
        return True

    def delete_playbook(self, playbook_id: str) -> bool:
        """Delete one playbook if it exists."""
        playbooks = self.load_playbooks()
        if playbook_id not in playbooks:
            return False
        del playbooks[playbook_id]
        self.save_playbooks(playbooks)
        return True

    def get_stats(self) -> dict[str, Any]:
        """Return aggregated storage statistics."""
        rules = self.load_rules()
        playbooks = self.load_playbooks()
        events = self.get_events(limit=1_000_000)

        return {
            "total_rules": len(rules),
            "total_playbooks": len(playbooks),
            "total_policy_events": len(events),
            "rules_file_size": self.rules_file.stat().st_size if self.rules_file.exists() else 0,
            "playbooks_file_size": self.playbooks_file.stat().st_size
            if self.playbooks_file.exists()
            else 0,
            "policy_events_file_size": self.policy_events_file.stat().st_size
            if self.policy_events_file.exists()
            else 0,
        }



#!/usr/bin/env python3
"""Public API entry points for intelligent confirmation."""

from __future__ import annotations

from pathlib import Path
from typing import Any

from .confirmation import IntelligentConfirmation
from .policy_store import PolicyStore


class ConfirmationAPI:
    """API wrapper for external integrations (for example gateway services)."""

    def __init__(self, memory_dir: Path | None = None):
        if memory_dir is None:
            memory_dir = Path.cwd() / ".clawpolicy"

        self.memory_dir = Path(memory_dir)
        self.policy_store = self._bootstrap_policy_store(self.memory_dir)
        self.conf_engine = IntelligentConfirmation(
            self.policy_store,
            project_scope_key=str(self.memory_dir.parent.resolve()),
        )

    @staticmethod
    def _bootstrap_policy_store(memory_dir: Path) -> PolicyStore:
        """Create and initialize policy storage files when missing."""
        return PolicyStore.bootstrap(memory_dir, ensure_files=True)

    def should_auto_execute(self, task: dict[str, Any]) -> tuple[bool, str, dict[str, Any]]:
        """Return `(auto_execute, reason, details)` for one task context."""
        decision = self.conf_engine.evaluate_task(task, persist=True)

        details: dict[str, Any] = {
            "decision_id": decision["decision_id"],
            "decision": decision,
            "matched_rules": decision["matched_rules"],
            "confidence": decision["confidence"]["max_confidence"],
            "success_streak": 0,
            "relevant_rules_count": decision["confidence"]["matched_rule_count"],
        }
        return decision["final_decision"] == "auto_execute", decision["reason"], details

    def record_execution_result(
        self,
        task: dict[str, Any],
        success: bool,
        auto_executed: bool,
        decision_id: str | None = None,
        execution_result: str | None = None,
        user_override: str | None = None,
    ) -> None:
        """Persist execution feedback into policy confidence history."""
        self.conf_engine.record_feedback(
            task,
            was_confirmed=not auto_executed,
            user_cancelled=not success,
            decision_id=decision_id,
            execution_result=execution_result,
            user_override=user_override,
        )

    def get_confidence_history(self, task_type: str | None = None) -> dict[str, Any]:
        """Return confidence history, optionally filtered by task type."""
        rules = self.policy_store.load_rules()

        if task_type:
            relevant_rules = [
                rule
                for rule in rules.values()
                if task_type in rule.trigger or f"task_type:{task_type}" in rule.trigger
            ]
        else:
            relevant_rules = [rule for rule in rules.values() if rule.confidence > 0.5]

        history = [
            {
                "id": rule.id,
                "summary": rule.summary,
                "confidence": rule.confidence,
                "success_streak": rule.success_streak,
                "status": rule.status,
                "scope": rule.scope,
                "source_type": rule.source_type,
                "policy_decision": rule.policy_decision,
            }
            for rule in sorted(relevant_rules, key=lambda item: -item.confidence)
        ]
        return {"rules": history}

    def get_recent_decisions(self, limit: int = 20) -> dict[str, Any]:
        """Return merged decision evaluation/outcome history."""
        events = list(reversed(self.policy_store.get_decision_events(limit=limit * 10)))
        merged: dict[str, dict[str, Any]] = {}

        for event in events:
            decision_id = event.payload.get("decision_id")
            if not decision_id:
                continue

            record = merged.setdefault(
                decision_id,
                {
                    "decision_id": decision_id,
                    "execution_result": "",
                    "user_override": "",
                    "rollback_happened": False,
                    "lifecycle_transition": "",
                },
            )
            if event.event_type == "decision_evaluated":
                record.update(event.payload)
            elif event.event_type == "decision_outcome":
                record["execution_result"] = event.payload.get("execution_result", "")
                record["user_override"] = event.payload.get("user_override", "")
                record["rollback_happened"] = event.payload.get("rollback_happened", False)
                record["lifecycle_transition"] = event.payload.get("lifecycle_transition", "")

        decisions = sorted(
            merged.values(),
            key=lambda item: item.get("timestamp", ""),
            reverse=True,
        )[:limit]
        return {"decisions": decisions}

    def get_explanation(self, task: dict[str, Any], should_confirm: bool, reason: str) -> str:
        """Return a readable explanation for one confirmation decision."""
        return self.conf_engine.get_explanation(task, should_confirm, reason)


def create_api(memory_dir: Path | None = None) -> ConfirmationAPI:
    """Convenience factory for `ConfirmationAPI`."""
    return ConfirmationAPI(memory_dir)



#!/usr/bin/env python3
"""
Experience replay buffer - Improve sample efficiency

Storage and sampling experience（state, action, reward, next_state, done），
Support priority sampling and batch sampling
"""

import numpy as np
from typing import List, Tuple
from dataclasses import dataclass
import random


@dataclass
class Experience:
    """single experience"""
    state: np.ndarray
    action: np.ndarray
    reward: float
    next_state: np.ndarray
    done: bool
    priority: float = 1.0  # priority（for priority sampling）

    def __repr__(self) -> str:
        return f"Experience(reward={self.reward:.3f}, done={self.done})"


class ExperienceReplay:
    """
    Experience replay buffer

    Function：
    - Store experience
    - random sampling
    - priority sampling（Optional）
    - Batch sampling
    """

    def __init__(self, capacity: int = 10000, use_prioritized: bool = False):
        """
        Initialize experience playback buffer

        Args:
            capacity: Buffer capacity
            use_prioritized: Whether to use priority sampling
        """
        self.capacity = capacity
        self.use_prioritized = use_prioritized

        # Buffer for storing experience
        self.buffer: List[Experience] = []
        self.position = 0

        # Priority sampling related
        self.priorities = np.zeros(capacity)
        self.max_priority = 1.0
        self.alpha = 0.6  # priority index
        self.beta = 0.4   # importance sampling index
        self.epsilon = 1e-6  # Avoid zero priority

    def add(self, experience: Experience) -> None:
        """
        Add experience to buffer

        Args:
            experience: Experience to add
        """
        if len(self.buffer) < self.capacity:
            self.buffer.append(experience)
            idx = len(self.buffer) - 1
        else:
            # Overwrite the oldest experience（circular buffer）
            idx = self.position
            self.buffer[idx] = experience

        if self.use_prioritized:
            self.priorities[idx] = self.max_priority

        self.position = (self.position + 1) % self.capacity

        # Update maximum priority
        if self.use_prioritized:
            self.max_priority = max(self.max_priority, experience.priority)

    def sample(self, batch_size: int = 32) -> List[Experience]:
        """
        Randomly sample a batch of experiences

        Args:
            batch_size: batch size

        Returns:
            experience batch
        """
        if len(self.buffer) < batch_size:
            batch_size = len(self.buffer)

        if self.use_prioritized:
            return self._prioritized_sample(batch_size)
        else:
            return random.sample(self.buffer, batch_size)

    def _prioritized_sample(self, batch_size: int) -> List[Experience]:
        """
        priority sampling

        Sampling based on priority，The higher the priority, the greater the probability of being sampled.
        """
        if len(self.buffer) == 0:
            return []

        # Calculate sampling probability
        priorities = self.priorities[:len(self.buffer)]
        probs = priorities ** self.alpha
        total = probs.sum()
        if total == 0:
            probs = np.ones_like(probs) / len(probs)
        else:
            probs /= total

        # Sampling index
        indices = np.random.choice(len(self.buffer), size=min(batch_size, len(self.buffer)), p=probs)

        # Calculate importance weight
        weights = (len(self.buffer) * probs[indices]) ** (-self.beta)
        weights /= weights.max()

        return [self.buffer[idx] for idx in indices]

    def get_batch(self, batch_size: int = 32) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Get batch data（for training）

        Args:
            batch_size: batch size

        Returns:
            (states, actions, rewards, next_states, dones)
        """
        experiences = self.sample(batch_size)

        if not experiences:
            return (
                np.zeros((0, 17)),  # states
                np.zeros((0, 4), dtype=int),  # actions (indices)
                np.zeros(0),      # rewards
                np.zeros((0, 17)), # next_states
                np.zeros(0)       # dones
            )

        states = np.array([exp.state for exp in experiences])
        actions = np.array([exp.action for exp in experiences])
        rewards = np.array([exp.reward for exp in experiences])
        next_states = np.array([exp.next_state for exp in experiences])
        dones = np.array([exp.done for exp in experiences])

        return states, actions, rewards, next_states, dones

    def update_priorities(self, indices: List[int], priorities: List[float]) -> None:
        """
        Update experience priority

        Args:
            indices: experience index
            priorities: new priority
        """
        for idx, priority in zip(indices, priorities):
            if idx < len(self.buffer):
                self.priorities[idx] = priority + self.epsilon
                self.max_priority = max(self.max_priority, priority)

    def __len__(self) -> int:
        return len(self.buffer)

    def is_ready(self, min_size: int = 100) -> bool:
        """
        Check if buffer is ready for sampling

        Args:
            min_size: Minimum size requirements

        Returns:
            Is it possible to sample
        """
        return len(self.buffer) >= min_size

    def clear(self) -> None:
        """Clear buffer"""
        self.buffer.clear()
        self.priorities = np.zeros(self.capacity)
        self.max_priority = 1.0
        self.position = 0


def main():
    """Test experience replay"""
    # Create experience replay buffer
    replay = ExperienceReplay(capacity=1000, use_prioritized=True)

    print(f"✅ Experience replay buffer created（capacity：{replay.capacity}）")

    # Add some simulation experience
    for i in range(10):
        state = np.random.randn(17)
        action = np.random.randint(0, 3, size=4)
        reward = np.random.randn()
        next_state = np.random.randn(17)
        done = i % 5 == 0

        exp = Experience(state, action, reward, next_state, done, priority=abs(reward))
        replay.add(exp)

    print(f"✅ Added {len(replay)} experience")

    # Sampling batch
    states, actions, rewards, next_states, dones = replay.get_batch(batch_size=4)

    print(f"✅ Sampling batch: states={states.shape}, actions={actions.shape}")
    print(f"   award: {rewards}")
    print(f"   Finish: {dones}")

    # Check if you are ready
    print(f"✅ Prepare for sampling: {replay.is_ready(min_size=5)}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Cross-platform installer for ClawPolicy."""

from __future__ import annotations

import argparse
import subprocess
import sys
from pathlib import Path


PROJECT_ROOT = Path(__file__).resolve().parent.parent


def run(cmd: list[str]) -> None:
    print("+", " ".join(cmd))
    subprocess.run(cmd, check=True)


def main() -> int:
    parser = argparse.ArgumentParser(description="Install ClawPolicy")
    parser.add_argument("--phase3", action="store_true", help="Install optional phase3 dependencies")
    parser.add_argument("--dev", action="store_true", help="Install development dependencies")
    parser.add_argument("--editable", action="store_true", help="Install package in editable mode")
    args = parser.parse_args()

    run([sys.executable, "-m", "pip", "install", "-r", str(PROJECT_ROOT / "requirements.txt")])
    if args.phase3:
        run([sys.executable, "-m", "pip", "install", "-r", str(PROJECT_ROOT / "requirements-full.txt")])
    if args.dev:
        run([sys.executable, "-m", "pip", "install", "-r", str(PROJECT_ROOT / "requirements-dev.txt")])

    if args.editable:
        run([sys.executable, "-m", "pip", "install", "-e", str(PROJECT_ROOT)])
    else:
        run([sys.executable, "-m", "pip", "install", str(PROJECT_ROOT)])

    print("Installation complete.")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



#!/usr/bin/env python3
"""Confirmation policy engine with structured explanations and feedback loops."""

from __future__ import annotations

from datetime import datetime, timezone
from typing import Any, Mapping, TypedDict
from uuid import uuid4

from .demotion import DemotionGate
from .explanation import render_explanation
from .policy_models import PolicyEvent, Rule
from .policy_store import PolicyStore
from .policy_resolution import (
    PolicyResolver,
    infer_domain,
    normalize_text,
    safe_files,
    safe_text,
    task_summary,
)
from .promotion import PromotionGate
from .risk import RiskAssessor, RiskLevel


class MatchedRuleInfo(TypedDict):
    """Serialized matched rule metadata."""

    id: str
    summary: str
    status: str
    scope: str
    scope_key: str
    source_type: str
    evidence_count: int
    policy_decision: str
    confidence: float
    success_streak: int
    last_seen_at: str


class ConfidenceInfo(TypedDict):
    """Confidence summary for rules related to one task."""

    max_confidence: float
    avg_confidence: float
    min_success_streak: int
    count: int
    rules: list[Rule]
    status_counts: dict[str, int]


class DecisionExplanation(TypedDict):
    """Structured explanation for a confirmation decision."""

    heuristic_basis: list[str]
    matched_rules: list[MatchedRuleInfo]
    resolution: str
    final_decision: str
    confidence: dict[str, Any]
    fallback_reason: str


class DecisionRecord(TypedDict):
    """Structured decision record used by API and audit logging."""

    decision_id: str
    task_summary: str
    task_description: str
    task_type: str
    command: str
    files: list[str]
    risk_level: str
    heuristic_basis: list[str]
    matched_rules: list[MatchedRuleInfo]
    final_decision: str
    reason: str
    resolution: str
    confidence: dict[str, Any]
    fallback_reason: str
    explanation: DecisionExplanation
    explanation_text: str
    timestamp: str
    domain: str
    scope: str
    scope_key: str


class IntelligentConfirmation:
    """Risk-aware confirmation policy with explicit explanation and audit hooks."""

    def __init__(
        self,
        policy_store: PolicyStore | None = None,
        project_scope_key: str = "",
    ) -> None:
        self.policy_store = policy_store
        self.project_scope_key = project_scope_key
        self.promotion_gate = PromotionGate(policy_store) if policy_store else None
        self.demotion_gate = DemotionGate()
        self.risk_assessor = RiskAssessor()
        self.policy_resolver = PolicyResolver(project_scope_key)

    @staticmethod
    def _now() -> str:
        return datetime.now(timezone.utc).isoformat()

    def should_confirm(self, task_context: Mapping[str, Any]) -> tuple[bool, str]:
        """Return `(needs_confirmation, reason)` for a task."""
        decision = self.evaluate_task(task_context, persist=False)
        return decision["final_decision"] == "require_confirmation", decision["reason"]

    def evaluate_task(self, task_context: Mapping[str, Any], persist: bool = False) -> DecisionRecord:
        """Build a structured confirmation decision for one task."""
        risk_level, heuristic_basis = self.assess_risk_details(task_context)
        scope, scope_key = self._infer_scope(task_context)
        domain = infer_domain(task_context)
        matched_rules = self._get_matched_rules(task_context, scope_key)
        confirmed_rules = [rule for rule in matched_rules if rule.status == "confirmed" and rule.policy_decision]
        selected_rule, resolution = self._resolve_rules(confirmed_rules)
        fallback_reason = ""

        if risk_level == RiskLevel.CRITICAL:
            final_decision = "require_confirmation"
            reason = "Critical risk heuristic requires confirmation"
            if matched_rules:
                fallback_reason = "heuristic override"
            resolution = resolution if resolution != "heuristic_default" else "critical_risk_override"
        elif selected_rule:
            final_decision = selected_rule.policy_decision
            if final_decision == "auto_execute":
                reason = f"Confirmed {selected_rule.scope} rule allows auto execution"
            else:
                reason = f"Confirmed {selected_rule.scope} rule requires confirmation"
        elif confirmed_rules:
            final_decision = "require_confirmation"
            fallback_reason = "ambiguous confirmed conflict"
            reason = "Conflicting confirmed rules require confirmation"
        else:
            if risk_level == RiskLevel.LOW:
                final_decision = "auto_execute"
                fallback_reason = "first-time task; heuristic fallback"
                reason = "Low-risk heuristic fallback allows auto execution"
            else:
                final_decision = "require_confirmation"
                if any(rule.status == "suspended" for rule in matched_rules):
                    fallback_reason = "suspended rule requires confirmation"
                else:
                    fallback_reason = "insufficient confidence" if matched_rules else "first-time task"
                reason = "No confirmed rule; confirmation required"

        confidence_info = self.get_confidence_info(task_context)
        serialized_rules = [self._serialize_rule(rule) for rule in matched_rules if rule.policy_decision]
        decision_scope = selected_rule.scope if selected_rule else scope
        decision_scope_key = selected_rule.scope_key if selected_rule else scope_key
        decision: DecisionRecord = {
            "decision_id": f"dec_{uuid4().hex[:12]}",
            "task_summary": self._task_summary(task_context),
            "task_description": safe_text(task_context, "task_description"),
            "task_type": safe_text(task_context, "task_type") or "T2",
            "command": safe_text(task_context, "command"),
            "files": safe_files(task_context),
            "risk_level": risk_level.value,
            "heuristic_basis": heuristic_basis,
            "matched_rules": serialized_rules,
            "final_decision": final_decision,
            "reason": reason,
            "resolution": resolution,
            "confidence": {
                "max_confidence": confidence_info["max_confidence"],
                "avg_confidence": confidence_info["avg_confidence"],
                "matched_rule_count": confidence_info["count"],
                "status_counts": confidence_info["status_counts"],
            },
            "fallback_reason": fallback_reason,
            "explanation": {
                "heuristic_basis": heuristic_basis,
                "matched_rules": serialized_rules,
                "resolution": resolution,
                "final_decision": final_decision,
                "confidence": {
                    "max_confidence": confidence_info["max_confidence"],
                    "avg_confidence": confidence_info["avg_confidence"],
                    "matched_rule_count": confidence_info["count"],
                    "status_counts": confidence_info["status_counts"],
                },
                "fallback_reason": fallback_reason,
            },
            "explanation_text": "",
            "timestamp": self._now(),
            "domain": domain,
            "scope": decision_scope,
            "scope_key": decision_scope_key,
        }
        decision["explanation_text"] = self.render_explanation(decision)

        if persist and self.policy_store:
            self._mark_rules_applied(matched_rules, decision["timestamp"])
            self._append_decision_event(decision)

        return decision

    def assess_risk(self, task_context: Mapping[str, Any]) -> RiskLevel:
        """Assess task risk from command, description and file paths."""
        return self.risk_assessor.assess(task_context)

    def assess_risk_details(self, task_context: Mapping[str, Any]) -> tuple[RiskLevel, list[str]]:
        """Assess risk and return basis used for the decision."""
        return self.risk_assessor.assess_details(task_context)

    def get_confidence_info(self, task_context: Mapping[str, Any]) -> ConfidenceInfo:
        """Return confidence metrics for rules matching the task."""
        empty_info: ConfidenceInfo = {
            "max_confidence": 0.0,
            "avg_confidence": 0.0,
            "min_success_streak": 0,
            "count": 0,
            "rules": [],
            "status_counts": {},
        }
        if not self.policy_store:
            return empty_info

        scope_key = self._infer_scope(task_context)[1]
        rules = [rule for rule in self.policy_store.load_rules().values() if self._matches_rule(rule, task_context, scope_key)]
        if not rules:
            return empty_info

        confidences = [rule.confidence for rule in rules]
        status_counts: dict[str, int] = {}
        for rule in rules:
            status_counts[rule.status] = status_counts.get(rule.status, 0) + 1

        return {
            "max_confidence": max(confidences),
            "avg_confidence": sum(confidences) / len(confidences),
            "min_success_streak": min(rule.success_streak for rule in rules),
            "count": len(rules),
            "rules": rules,
            "status_counts": status_counts,
        }

    def record_feedback(
        self,
        task_context: Mapping[str, Any],
        was_confirmed: bool,
        user_cancelled: bool,
        decision_id: str | None = None,
        execution_result: str | None = None,
        user_override: str | None = None,
    ) -> None:
        """Update matching rules and audit events from task execution feedback."""
        if not self.policy_store:
            return

        decision = self._find_decision(decision_id)
        if decision is None:
            decision = self.evaluate_task(task_context, persist=True)

        outcome = execution_result or ("cancel" if user_cancelled else "success")
        matched_rule_ids = [rule["id"] for rule in decision["matched_rules"]]
        rollback_happened = outcome == "rollback"
        lifecycle_transitions: list[str] = []

        rules = self.policy_store.load_rules()
        evidence_type = self._classify_evidence_source(user_override, decision["final_decision"], outcome)
        desired_policy = self._desired_policy_decision(user_override, decision["final_decision"], outcome)

        target_rules = [
            rules[rule_id]
            for rule_id in matched_rule_ids
            if rule_id in rules and rules[rule_id].policy_decision == desired_policy
        ]
        if evidence_type in {"repeated_failure"} or user_override in {
            "prefer_confirmation",
            "blocked_auto_execute",
            "should_have_required_confirmation",
        }:
            target_rules = [rules[rule_id] for rule_id in matched_rule_ids if rule_id in rules]

        if evidence_type in {
            "explicit_correction",
            "explicit_preference",
            "repeated_failure",
            "missed_confirmation",
        } and not target_rules:
            new_rule = self._create_candidate_rule(task_context, desired_policy, evidence_type)
            new_rule.risk_level = decision["risk_level"]
            rules[new_rule.id] = new_rule
            target_rules = [new_rule]

        updated = False
        for rule in target_rules:
            previous_status = rule.status
            if not self._apply_feedback_to_rule(rule, outcome, evidence_type, user_override, decision["risk_level"]):
                continue

            has_conflict = self._has_unresolved_conflict(rule, rules)
            transition = None
            if previous_status == "confirmed":
                suspend_trigger = self._suspension_trigger(user_override, decision["final_decision"], outcome)
                transition = self.demotion_gate.maybe_suspend(
                    rule,
                    trigger=suspend_trigger,
                    has_conflict=has_conflict,
                )
            elif previous_status == "suspended":
                transition = self.demotion_gate.maybe_reactivate(rule)

            if transition is None and rule.status == "candidate" and self.promotion_gate:
                promotion = self.promotion_gate.maybe_promote(rule, decision["risk_level"], has_conflict)
                if promotion.promoted:
                    transition = self._make_transition("rule_promoted", "candidate_promoted")

            if transition is None and rule.status == "suspended":
                transition = self.demotion_gate.maybe_archive(
                    rule,
                    superseded=self._is_superseded(rule, rules),
                )

            rule.last_lifecycle_event_at = self._now() if transition else rule.last_lifecycle_event_at
            rule.calculate_asset_id()
            self._append_rule_event(rule, decision, evidence_type, transition)
            if transition is not None:
                lifecycle_transitions.append(self._transition_event_type(transition))
            updated = True

        if updated:
            self._archive_superseded_suspended_rules(rules, decision)
        self.policy_store.save_rules(rules)
        self.policy_store.append_event(
            PolicyEvent(
                timestamp=self._now(),
                event_type="decision_outcome",
                asset_id="",
                trigger_signals=[decision["task_type"]],
                changes=f"Decision outcome: {outcome}",
                source_node_id="confirmation_policy",
                payload={
                    "decision_id": decision["decision_id"],
                    "final_decision": decision["final_decision"],
                    "user_override": user_override or "",
                    "execution_result": outcome,
                    "was_confirmed": was_confirmed,
                    "rule_ids": matched_rule_ids,
                    "task_summary": decision["task_summary"],
                    "task_type": decision["task_type"],
                    "command": decision["command"],
                    "files": decision["files"],
                    "risk_level": decision["risk_level"],
                    "scope": decision["scope"],
                    "scope_key": decision["scope_key"],
                    "heuristic_basis": decision["heuristic_basis"],
                    "matched_rule_ids": matched_rule_ids,
                    "explanation": decision["explanation"],
                    "rollback_happened": rollback_happened,
                    "lifecycle_transition": ",".join(lifecycle_transitions),
                },
            )
        )

    def get_explanation(self, task_context: Mapping[str, Any], should_confirm: bool, reason: str) -> str:
        """Return a human-readable explanation for the decision."""
        _ = should_confirm, reason
        decision = self.evaluate_task(task_context, persist=False)
        return self.render_explanation(decision)

    def render_explanation(self, decision: DecisionRecord) -> str:
        """Render a structured decision as concise audit text."""
        return render_explanation(decision)

    def _task_summary(self, task_context: Mapping[str, Any]) -> str:
        return task_summary(task_context)

    def _infer_scope(self, task_context: Mapping[str, Any]) -> tuple[str, str]:
        return self.policy_resolver.infer_scope(task_context)

    def _get_matched_rules(self, task_context: Mapping[str, Any], scope_key: str) -> list[Rule]:
        if not self.policy_store:
            return []
        rules = self.policy_store.load_rules().values()
        return self.policy_resolver.get_matched_rules(rules, task_context, scope_key)

    def _matches_rule(self, rule: Rule, task_context: Mapping[str, Any], scope_key: str) -> bool:
        return self.policy_resolver.matches_rule(rule, task_context, scope_key)

    def _resolve_rules(self, rules: list[Rule]) -> tuple[Rule | None, str]:
        return self.policy_resolver.resolve_rules(rules)

    def _serialize_rule(self, rule: Rule) -> MatchedRuleInfo:
        return {
            "id": rule.id,
            "summary": rule.summary,
            "status": rule.status,
            "scope": rule.scope,
            "scope_key": rule.scope_key,
            "source_type": rule.source_type,
            "evidence_count": rule.evidence_count,
            "policy_decision": rule.policy_decision,
            "confidence": rule.confidence,
            "success_streak": rule.success_streak,
            "last_seen_at": rule.last_seen_at,
        }

    def _mark_rules_applied(self, matched_rules: list[Rule], timestamp: str) -> None:
        if not self.policy_store or not matched_rules:
            return
        rules = self.policy_store.load_rules()
        updated = False
        for rule in matched_rules:
            if rule.id not in rules:
                continue
            rules[rule.id].last_applied_at = timestamp
            rules[rule.id].last_seen_at = timestamp or rules[rule.id].last_seen_at
            updated = True
        if updated:
            self.policy_store.save_rules(rules)

    def _append_decision_event(self, decision: DecisionRecord) -> None:
        if not self.policy_store:
            return
        self.policy_store.append_event(
            PolicyEvent(
                timestamp=decision["timestamp"],
                event_type="decision_evaluated",
                asset_id="",
                trigger_signals=[decision["task_type"]],
                changes=decision["reason"],
                source_node_id="confirmation_policy",
                payload=dict(decision),
            )
        )

    def _find_decision(self, decision_id: str | None) -> DecisionRecord | None:
        if not self.policy_store or not decision_id:
            return None
        for event in self.policy_store.get_decision_events(limit=1_000):
            if event.event_type != "decision_evaluated":
                continue
            if event.payload.get("decision_id") == decision_id:
                return event.payload  # type: ignore[return-value]
        return None

    def _classify_evidence_source(
        self,
        user_override: str | None,
        final_decision: str,
        execution_result: str,
    ) -> str:
        if user_override in {"prefer_auto_execute", "prefer_confirmation"}:
            return "explicit_preference"
        if user_override in {"confirmed_after_prompt", "blocked_auto_execute", "forced_auto_execute"}:
            return "explicit_correction"
        if user_override == "should_have_required_confirmation":
            return "missed_confirmation"
        if execution_result in {"failure", "rollback"} and final_decision == "auto_execute":
            return "repeated_failure"
        if execution_result == "success" and final_decision == "auto_execute":
            return "repeated_success"
        return ""

    def _suspension_trigger(
        self,
        user_override: str | None,
        final_decision: str,
        execution_result: str,
    ) -> str:
        if execution_result == "rollback":
            return "rollback"
        if user_override in {"prefer_confirmation", "blocked_auto_execute"}:
            return "override"
        if execution_result in {"failure"} and final_decision == "auto_execute":
            return "failure"
        if user_override == "should_have_required_confirmation":
            return "missed_confirmation"
        return ""

    def _desired_policy_decision(
        self,
        user_override: str | None,
        final_decision: str,
        execution_result: str,
    ) -> str:
        if user_override in {"prefer_auto_execute", "confirmed_after_prompt", "forced_auto_execute"}:
            return "auto_execute"
        if user_override in {"prefer_confirmation", "blocked_auto_execute"}:
            return "require_confirmation"
        if execution_result in {"failure", "rollback"} and final_decision == "auto_execute":
            return "require_confirmation"
        return final_decision

    def _create_candidate_rule(
        self,
        task_context: Mapping[str, Any],
        policy_decision: str,
        source_type: str,
    ) -> Rule:
        scope, scope_key = self._infer_scope(task_context)
        summary = (
            f"{'Auto execute' if policy_decision == 'auto_execute' else 'Require confirmation'} "
            f"for {self._task_summary(task_context)}"
        )
        rule = Rule(
            id=f"policy_{uuid4().hex[:10]}",
            summary=summary,
            category="harden",
            trigger=self._build_rule_triggers(task_context),
            strategy=self._build_rule_strategy(task_context, policy_decision),
            validation=[self._build_validation_assertion(task_context, policy_decision)],
            confidence=0.6 if source_type.startswith("explicit") else 0.5,
            success_streak=0,
            status="candidate",
            scope=scope,
            scope_key=scope_key,
            evidence_count=1,
            source_type=source_type,
            last_seen_at=self._now(),
            last_applied_at="",
            policy_decision=policy_decision,
            risk_level=self.assess_risk(task_context).value,
        )
        rule.calculate_asset_id()
        return rule

    def _build_rule_triggers(self, task_context: Mapping[str, Any]) -> list[str]:
        triggers: list[str] = []
        task_type = safe_text(task_context, "task_type") or "T2"
        triggers.append(f"task_type:{task_type}")
        command = normalize_text(safe_text(task_context, "command"))
        if command:
            triggers.append(f"command:{command}")
        else:
            triggers.append(f"task:{normalize_text(self._task_summary(task_context))}")
        return triggers

    def _build_rule_strategy(self, task_context: Mapping[str, Any], policy_decision: str) -> str:
        summary = self._task_summary(task_context)
        if policy_decision == "auto_execute":
            return f"Prefer automatic execution for tasks matching: {summary}"
        return f"Require explicit confirmation for tasks matching: {summary}"

    def _build_validation_assertion(self, task_context: Mapping[str, Any], policy_decision: str) -> str:
        task_type = safe_text(task_context, "task_type") or "T2"
        command = safe_text(task_context, "command") or self._task_summary(task_context)
        return (
            "assert decision.final_decision == "
            f"'{policy_decision}' for task_type={task_type} command={command}"
        )

    def _apply_feedback_to_rule(
        self,
        rule: Rule,
        outcome: str,
        evidence_type: str,
        user_override: str | None,
        risk_level: str,
    ) -> bool:
        if not evidence_type:
            return False

        timestamp = self._now()
        if rule.status == "hint" and evidence_type not in {"git_history", "rl_feedback", "legacy"}:
            rule.status = "candidate"

        rule.last_seen_at = timestamp
        rule.evidence_count += 1
        rule.risk_level = risk_level or rule.risk_level
        if not rule.last_applied_at:
            rule.last_applied_at = timestamp

        if user_override == "should_have_required_confirmation":
            rule.conflict_count += 1

        if outcome == "success":
            rule.success_streak += 1
            rule.failure_streak = 0
            rule.confidence = min(1.0, rule.confidence + (0.1 if evidence_type.startswith("explicit") else 0.05))
            if rule.policy_decision == "auto_execute":
                rule.accepted_auto_execute_count += 1
            if user_override in {"confirmed_after_prompt", "prefer_auto_execute", "forced_auto_execute"}:
                rule.accepted_confirmation_count += 1
            if user_override in {"prefer_confirmation", "blocked_auto_execute"}:
                rule.override_streak += 1
            else:
                rule.override_streak = 0
        else:
            rule.success_streak = 0
            if rule.policy_decision == "auto_execute":
                rule.confidence = max(0.0, rule.confidence - 0.2)
            else:
                rule.confidence = min(1.0, rule.confidence + 0.05)
            if outcome in {"failure", "rollback"}:
                rule.failure_streak += 1
            if outcome == "rollback":
                rule.rollback_count += 1
            if user_override in {"prefer_confirmation", "blocked_auto_execute"}:
                rule.override_streak += 1
        return True

    def _has_unresolved_conflict(self, rule: Rule, rules: Mapping[str, Rule]) -> bool:
        rule_signature = tuple(sorted(rule.trigger))
        for other in rules.values():
            if other.id == rule.id or other.status != "confirmed":
                continue
            if other.scope != rule.scope or other.scope_key != rule.scope_key:
                continue
            if tuple(sorted(other.trigger)) != rule_signature:
                continue
            if other.policy_decision != rule.policy_decision:
                return True
        return False

    def _is_superseded(self, rule: Rule, rules: Mapping[str, Rule]) -> bool:
        rule_signature = tuple(sorted(rule.trigger))
        for other in rules.values():
            if other.id == rule.id or other.status != "confirmed":
                continue
            if other.scope != rule.scope or other.scope_key != rule.scope_key:
                continue
            if tuple(sorted(other.trigger)) != rule_signature:
                continue
            return True
        return False

    def _make_transition(self, event_type: str, reason: str) -> dict[str, str]:
        return {"event_type": event_type, "reason": reason}

    @staticmethod
    def _transition_event_type(transition: Mapping[str, str] | Any) -> str:
        if hasattr(transition, "transition"):
            return str(transition.transition)
        return str(transition["event_type"])

    @staticmethod
    def _transition_reason(transition: Mapping[str, str] | Any) -> str:
        if hasattr(transition, "reason"):
            return str(transition.reason)
        return str(transition["reason"])

    def _append_rule_event(
        self,
        rule: Rule,
        decision: DecisionRecord,
        source_type: str,
        transition: Mapping[str, str] | Any | None,
    ) -> None:
        if not self.policy_store:
            return

        event_type = self._transition_event_type(transition) if transition else "rule_updated"
        reason = self._transition_reason(transition) if transition else "runtime_feedback"
        self.policy_store.append_event(
            PolicyEvent(
                timestamp=self._now(),
                event_type=event_type,
                asset_id=rule.asset_id,
                trigger_signals=[decision["task_type"]],
                changes=f"Policy rule updated: {rule.status} / {rule.policy_decision}",
                source_node_id=source_type or "runtime_feedback",
                payload={
                    "rule_id": rule.id,
                    "status": rule.status,
                    "policy_decision": rule.policy_decision,
                    "evidence_count": rule.evidence_count,
                    "trigger": reason,
                    "scope": rule.scope,
                    "scope_key": rule.scope_key,
                    "risk_level": rule.risk_level,
                },
            )
        )

    def _archive_superseded_suspended_rules(
        self,
        rules: Mapping[str, Rule],
        decision: DecisionRecord,
    ) -> None:
        if not self.policy_store:
            return
        for rule in rules.values():
            if rule.status != "suspended":
                continue
            transition = self.demotion_gate.maybe_archive(rule, superseded=self._is_superseded(rule, rules))
            if transition is None:
                continue
            rule.last_lifecycle_event_at = self._now()
            rule.calculate_asset_id()
            self._append_rule_event(rule, decision, "lifecycle", transition)



#!/usr/bin/env python3
"""Lifecycle demotion gates for autonomous confirmation policies."""

from __future__ import annotations

from dataclasses import dataclass

from .policy_models import Rule

SUSPEND_FAILURE_STREAK = 2
SUSPEND_OVERRIDE_STREAK = 2
SUSPEND_MISSED_CONFIRMATION_COUNT = 2
REACTIVATE_POSITIVE_EVIDENCE = 2
ARCHIVE_NEGATIVE_EVIDENCE = 3


@dataclass
class DemotionResult:
    """Structured suspension/archive/reactivation decision."""

    transition: str
    reason: str


class DemotionGate:
    """Evaluate when policies should be suspended, reactivated, or archived."""

    def maybe_suspend(
        self,
        rule: Rule,
        *,
        trigger: str,
        has_conflict: bool = False,
    ) -> DemotionResult | None:
        if rule.status != "confirmed":
            return None

        if trigger == "rollback":
            rule.status = "suspended"
            rule.suspension_reason = "rollback"
            return DemotionResult("rule_suspended", "rollback")
        if has_conflict:
            rule.status = "suspended"
            rule.suspension_reason = "conflict"
            return DemotionResult("rule_suspended", "conflict")
        if trigger == "failure" and rule.failure_streak >= SUSPEND_FAILURE_STREAK:
            rule.status = "suspended"
            rule.suspension_reason = "repeated_failure"
            return DemotionResult("rule_suspended", "repeated_failure")
        if trigger == "override" and rule.override_streak >= SUSPEND_OVERRIDE_STREAK:
            rule.status = "suspended"
            rule.suspension_reason = "repeated_override"
            return DemotionResult("rule_suspended", "repeated_override")
        if trigger == "missed_confirmation" and rule.conflict_count >= SUSPEND_MISSED_CONFIRMATION_COUNT:
            rule.status = "suspended"
            rule.suspension_reason = "missed_confirmation"
            return DemotionResult("rule_suspended", "missed_confirmation")
        return None

    def maybe_reactivate(self, rule: Rule) -> DemotionResult | None:
        if rule.status != "suspended":
            return None
        if (
            rule.accepted_confirmation_count >= REACTIVATE_POSITIVE_EVIDENCE
            or rule.accepted_auto_execute_count >= REACTIVATE_POSITIVE_EVIDENCE
        ):
            rule.status = "candidate"
            rule.suspension_reason = ""
            rule.failure_streak = 0
            rule.override_streak = 0
            rule.conflict_count = 0
            return DemotionResult("rule_reactivated", "positive_evidence")
        return None

    def maybe_archive(self, rule: Rule, *, superseded: bool = False) -> DemotionResult | None:
        if rule.status != "suspended":
            return None
        negative_score = rule.failure_streak + rule.override_streak + rule.rollback_count + rule.conflict_count
        if superseded:
            rule.status = "archived"
            return DemotionResult("rule_archived", "superseded")
        if negative_score >= ARCHIVE_NEGATIVE_EVIDENCE:
            rule.status = "archived"
            return DemotionResult("rule_archived", "continued_negative_evidence")
        return None



#!/usr/bin/env python3
"""
Core dimensions and enumeration order contract

Centralized maintenance status/action space dimensions，Avoid multi-file hardcoded drift。
"""

from typing import Final


TASK_TYPE_ORDER: Final[tuple[str, ...]] = ("T1", "T2", "T3", "T4")
TECH_STACK_ORDER: Final[tuple[str, ...]] = (
    "react",
    "vue",
    "fastapi",
    "express",
    "python",
    "javascript",
    "typescript",
    "go",
)
USER_MOOD_ORDER: Final[tuple[str, ...]] = ("focused", "relaxed", "stressed")

AGENT_ORDER: Final[tuple[str, ...]] = ("claude", "codex", "gemini")
AUTOMATION_ORDER: Final[tuple[str, ...]] = ("low", "medium", "high")
STYLE_ORDER: Final[tuple[str, ...]] = ("brief", "detailed", "interactive")
CONFIRM_ORDER: Final[tuple[bool, ...]] = (False, True)

STATE_DIMENSIONS: Final[dict[str, int]] = {
    "task_type": len(TASK_TYPE_ORDER),
    "tech_stack": len(TECH_STACK_ORDER),
    "user_mood": len(USER_MOOD_ORDER),
    "time_of_day": 1,
    "recent_performance": 1,
}
STATE_VECTOR_DIM: Final[int] = sum(STATE_DIMENSIONS.values())

ACTION_HEAD_DIMS: Final[dict[str, int]] = {
    "agent": len(AGENT_ORDER),
    "automation": len(AUTOMATION_ORDER),
    "style": len(STYLE_ORDER),
    "confirm": len(CONFIRM_ORDER),
}
ACTION_VECTOR_DIM: Final[int] = sum(ACTION_HEAD_DIMS.values())



#!/usr/bin/env python3
"""Canonical policy model and store tests."""

from __future__ import annotations

from pathlib import Path

from lib.policy_models import Playbook, PolicyEvent, Rule
from lib.policy_store import PolicyStore


def test_policy_models_use_canonical_type_names() -> None:
    rule = Rule(id="rule_test", summary="Run tests safely")
    playbook = Playbook(id="playbook_test", summary="Safe test playbook")
    event = PolicyEvent(
        timestamp="2026-03-01T12:00:00Z",
        event_type="decision_evaluated",
        asset_id="sha256:test",
    )

    assert rule.type == "Rule"
    assert playbook.type == "Playbook"
    assert event.type == "PolicyEvent"


def test_policy_store_uses_canonical_files(tmp_path: Path) -> None:
    store = PolicyStore(tmp_path)

    assert store.rules_file == tmp_path / "rules.json"
    assert store.playbooks_file == tmp_path / "playbooks.json"
    assert store.policy_events_file == tmp_path / "policy_events.jsonl"


def test_policy_store_bootstrap_provisions_canonical_files(tmp_path: Path) -> None:
    memory_dir = tmp_path / ".clawpolicy"

    store = PolicyStore.bootstrap(memory_dir, ensure_files=True)

    assert store.base_dir == memory_dir / "policy"
    assert store.rules_file.exists()
    assert store.playbooks_file.exists()
    assert store.policy_events_file.exists()


def test_policy_store_round_trips_rules_playbooks_and_events(tmp_path: Path) -> None:
    store = PolicyStore(tmp_path)
    rule = Rule(id="rule_test", summary="Run tests safely", confidence=0.8)
    playbook = Playbook(id="playbook_test", summary="Safe test playbook", rules_used=["rule_test"])
    event = PolicyEvent(
        timestamp="2026-03-01T12:00:00Z",
        event_type="decision_evaluated",
        asset_id="sha256:test",
        payload={"decision_id": "dec_123"},
    )

    store.save_rule(rule)
    store.save_playbook(playbook)
    store.append_event(event)

    assert store.get_rule("rule_test") is not None
    assert store.get_playbook("playbook_test") is not None
    assert store.get_policy_events(limit=5)[0].payload["decision_id"] == "dec_123"


def test_rule_from_dict_supports_suspended_and_domain_scope() -> None:
    rule = Rule.from_dict(
        {
            "id": "rule_test",
            "summary": "Use doc rules for docs tasks",
            "status": "suspended",
            "scope": "domain",
            "scope_key": "docs",
            "policy_decision": "require_confirmation",
        }
    )

    assert rule.status == "suspended"
    assert rule.scope == "domain"
    assert rule.scope_key == "docs"



#!/usr/bin/env python3
"""
Core Contract Conformance Testing
"""

from lib.agent import PolicyNetwork
from lib.contracts import ACTION_HEAD_DIMS, ACTION_VECTOR_DIM, STATE_VECTOR_DIM
from lib.environment import InteractionEnvironment


def test_action_head_dims_sum_matches_action_vector_dim():
    assert sum(ACTION_HEAD_DIMS.values()) == ACTION_VECTOR_DIM


def test_environment_space_sizes_match_contract():
    env = InteractionEnvironment()
    assert env.get_state_space_size() == STATE_VECTOR_DIM
    assert env.get_action_space_size() == ACTION_VECTOR_DIM


def test_policy_default_action_dim_matches_contract():
    policy = PolicyNetwork(state_dim=STATE_VECTOR_DIM)
    assert policy.action_dim == ACTION_VECTOR_DIM



#!/usr/bin/env python3
"""
Test intent alignment functionality
"""

import sys
from pathlib import Path

# Add project root directory to path
sys.path.insert(0, str(Path(__file__).parent.parent))

from lib.integration import IntentAlignmentEngine


def test_collector():
    """Test data collection"""
    print("="*50)
    print("test1: data collection")
    print("="*50)

    engine = IntentAlignmentEngine(".")
    git_data = engine.collector.collect(50)

    assert git_data is not None, "Data collection failed"
    assert "tech_stack" in git_data, "Missing technology stack data"
    assert "metadata" in git_data, "Missing metadata"

    print("✅ Data collection test passed")
    print()


def test_learner():
    """Test learning algorithms"""
    print("="*50)
    print("test2: learning algorithm")
    print("="*50)

    # createengine
    from lib.learner import PreferenceLearner
    learner = PreferenceLearner()

    # simulated data
    mock_data = {
        "tech_stack": {
            "python": 10,
            "javascript": 5,
            "react": 8
        },
        "workflow": {
            "test_first": True,
            "test_ratio": 0.4
        },
        "metadata": {
            "confidence": 0.8
        }
    }

    preferences = learner.learn_from_git_history(mock_data)

    assert preferences is not None, "learning failure"
    assert "tech_stack" in preferences, "Missing technology stack preference"
    assert preferences["tech_stack"]["primary"] == "python", "Major technical identification errors"

    print("✅ Learning algorithm test passed")
    print()


def test_integration():
    """Test full integration"""
    print("="*50)
    print("test3: Full integration")
    print("="*50)

    engine = IntentAlignmentEngine(".")

    # Run analysis
    result = engine.run_analysis(10)

    assert result is not None, "Analysis failed"
    assert "tech_stack" in result, "Missing technology stack results"

    print("✅ Integration test passed")
    print()


def main():
    """Run all tests"""
    print("\n🧪 Start testing the intent alignment feature...\n")

    try:
        test_collector()
        test_learner()
        # test_integration()  # needGitstorehouse，Skip for now

        print("="*50)
        print("✅ All tests passed！")
        print("="*50)

    except AssertionError as e:
        print(f"❌ test failed: {e}")
        sys.exit(1)
    except Exception as e:
        print(f"❌ mistake: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Experience replay tests."""

import numpy as np
from lib.experience_replay import ExperienceReplay, Experience


def _make_exp(reward: float) -> Experience:
    return Experience(
        state=np.zeros(17),
        action=np.zeros(4),
        reward=reward,
        next_state=np.zeros(17),
        done=False,
        priority=abs(reward)
    )


def test_ring_buffer_overwrite():
    replay = ExperienceReplay(capacity=3, use_prioritized=False)

    for i in range(5):
        replay.add(_make_exp(float(i)))

    assert len(replay) == 3

    rewards = [exp.reward for exp in replay.buffer]
    assert set(rewards) == {2.0, 3.0, 4.0}


if __name__ == "__main__":
    import pytest
    pytest.main([__file__, "-v"])



#!/usr/bin/env python3
"""
Reinforcement learning interactive environment - ClawPolicy interactive environment

Define state space and action space，accomplish ClawPolicy interactive OpenAI Gymstyle interface：
- state: task context（task_type, tech_stack, user_moodwait）
- action: Agentchoose（agent_selection, automation_levelwait）
- reward: fromreward.RewardCalculatorcalculate
- done: Is the task completed?

support：
- reset(task_context): Reset environment to new task
- step(action): perform action → (next_state, reward, done, info)
- _encode_state(context): Encode context into state vector
"""

import numpy as np
from typing import Dict, List, Any, Tuple, Optional
from dataclasses import dataclass
from enum import Enum
import json
from pathlib import Path

from .contracts import (
    ACTION_HEAD_DIMS,
    ACTION_VECTOR_DIM,
    AGENT_ORDER,
    AUTOMATION_ORDER,
    CONFIRM_ORDER,
    STATE_DIMENSIONS,
    STATE_VECTOR_DIM,
    STYLE_ORDER,
    TECH_STACK_ORDER,
    USER_MOOD_ORDER,
)
from .reward import RewardCalculator


class TaskType(Enum):
    """Task type enum"""
    T1 = "T1"  # lightweight：<20OK
    T2 = "T2"  # medium：20-200OK
    T3 = "T3"  # Severe：200+OK
    T4 = "T4"  # Danger：core system


class AgentType(Enum):
    """Agenttype enum"""
    CLAUDE = "claude"
    CODEX = "codex"
    GEMINI = "gemini"


class AutomationLevel(Enum):
    """Automation level enumeration"""
    LOW = "low"  # Requires frequent confirmation
    MEDIUM = "medium"  # Partially automated
    HIGH = "high"  # Highly automated


class CommunicationStyle(Enum):
    """communication style enum"""
    BRIEF = "brief"  # concise
    DETAILED = "detailed"  # detailed
    INTERACTIVE = "interactive"  # interactive


@dataclass
class State:
    """Status data class"""
    # Task type（one-hotcoding）
    task_type: np.ndarray  # [4] (T1, T2, T3, T4)

    # technology stack（one-hotcoding）
    tech_stack: np.ndarray  # [N] Based on the number of supported technologies

    # User mood（one-hotcoding）
    user_mood: np.ndarray  # [3] (focused, relaxed, stressed)

    # time（normalization0-1）
    time_of_day: float  # 0-1 (0=midnight, 0.5=noon, 1=midnight)

    # Recent performance（normalization0-1）
    recent_performance: float  # 0-1

    # historyAgentuse（statistics）
    agent_usage_history: Dict[str, int]  # agentname -> Number of uses

    def to_vector(self) -> np.ndarray:
        """Convert state to vector"""
        # Combine all features
        vectors: List[np.ndarray] = [
            self.task_type.flatten(),
            self.tech_stack.flatten(),
            self.user_mood.flatten(),
            np.array([self.time_of_day], dtype=float),
            np.array([self.recent_performance], dtype=float),
        ]

        return np.concatenate(vectors)

    def __repr__(self) -> str:
        return (f"State(task_type={self.task_type}, "
                f"tech_stack={self.tech_stack}, "
                f"user_mood={self.user_mood}, "
                f"time={self.time_of_day:.2f}, "
                f"perf={self.recent_performance:.2f})")


@dataclass
class Action:
    """action data class"""
    agent_selection: AgentType  # Which one to chooseAgent
    automation_level: AutomationLevel  # Automation level
    communication_style: CommunicationStyle  # communication style
    confirmation_needed: bool  # Do you need to confirm

    def to_vector(self, agent_map: Dict[AgentType, int],
                  automation_map: Dict[AutomationLevel, int],
                  style_map: Dict[CommunicationStyle, int],
                  confirm_map: Dict[bool, int]) -> np.ndarray:
        """Convert action toone-hotencoding vector"""
        vector_size = (
            len(agent_map) +
            len(automation_map) +
            len(style_map) +
            len(confirm_map)  # confirmation_needed
        )

        vector = np.zeros(vector_size)

        # Agentchoose（one-hot）
        vector[agent_map[self.agent_selection]] = 1

        # Automation level（one-hot）
        offset = len(agent_map)
        vector[offset + automation_map[self.automation_level]] = 1

        # communication style（one-hot）
        offset += len(automation_map)
        vector[offset + style_map[self.communication_style]] = 1

        # Confirmation sign
        offset += len(style_map)
        vector[offset + confirm_map[self.confirmation_needed]] = 1.0

        return vector

    def __repr__(self) -> str:
        return (f"Action(agent={self.agent_selection.value}, "
                f"automation={self.automation_level.value}, "
                f"style={self.communication_style.value}, "
                f"confirm={self.confirmation_needed})")


class InteractionEnvironment:
    """
    ClawPolicy interactive environment

    accomplishOpenAI Gymstyle environment interface：
    - observation_space: state space dimensions
    - action_space: action space dimensions
    - reset(): Reset environment
    - step(): perform action
    """

    # State space definition
    STATE_DIM = STATE_DIMENSIONS.copy()
    TOTAL_STATE_DIM = STATE_VECTOR_DIM  # 17dimension

    # action space definition
    ACTION_DIM = {
        "agent_selection": ACTION_HEAD_DIMS["agent"],
        "automation_level": ACTION_HEAD_DIMS["automation"],
        "communication_style": ACTION_HEAD_DIMS["style"],
        "confirmation_needed": ACTION_HEAD_DIMS["confirm"],
    }
    TOTAL_ACTION_DIM = ACTION_VECTOR_DIM  # 11dimension

    # Supported technology stack
    SUPPORTED_TECH = {name: idx for idx, name in enumerate(TECH_STACK_ORDER)}

    # Agentmapping
    AGENT_MAP = {AgentType(name): idx for idx, name in enumerate(AGENT_ORDER)}

    # Automation level mapping
    AUTOMATION_MAP = {AutomationLevel(name): idx for idx, name in enumerate(AUTOMATION_ORDER)}

    # communication style mapping
    STYLE_MAP = {CommunicationStyle(name): idx for idx, name in enumerate(STYLE_ORDER)}

    # Confirm flag mapping
    CONFIRM_MAP = {value: idx for idx, value in enumerate(CONFIRM_ORDER)}

    def __init__(self, config_path: Optional[str] = None):
        """
        Initialize environment

        Args:
            config_path: Model configuration path（Used to load historical data）
        """
        self.reward_calculator = RewardCalculator()

        # Current status
        self.current_state: Optional[State] = None
        self.current_task_context: Optional[Dict[str, Any]] = None

        # historical statistics
        self.episode_rewards: List[float] = []
        self.episode_count = 0
        self.agent_usage_history: Dict[str, int] = {
            "claude": 0,
            "codex": 0,
            "gemini": 0
        }

        # Recent performance（Initial is0.5Means medium）
        self.recent_performance: float = 0.5

        # Try loading history
        if config_path:
            self._load_history(config_path)

    def reset(self, task_context: Dict[str, Any]) -> State:
        """
        Reset environment to new task

        Args:
            task_context: task context，Include：
                - task_type: Task type（"T1", "T2", "T3", "T4"）
                - tech_stack: Technology stack used（list）
                - user_mood: User mood（"focused", "relaxed", "stressed"，Optional）
                - time_of_day: current time（0-24Hour，Optional）

        Returns:
            initial state
        """
        # Parse task type
        task_type_str = str(task_context.get("task_type", "T2")).upper()
        if task_type_str not in TaskType.__members__:
            task_type_str = "T2"
        task_type = TaskType[task_type_str]
        task_context_normalized = dict(task_context)
        task_context_normalized["task_type"] = task_type_str

        # Analyze technology stack
        tech_stacks = task_context.get("tech_stack", ["python"])
        tech_stack_vector = self._encode_tech_stack(tech_stacks)

        # Analyze user mood（Default isfocused）
        user_mood_str = task_context.get("user_mood", "focused")
        user_mood_vector = self._encode_user_mood(user_mood_str)

        # parsing time（Default is noon）
        time_of_day = task_context.get("time_of_day", 12.0)
        try:
            time_of_day = float(time_of_day)
        except (TypeError, ValueError):
            time_of_day = 12.0
        time_of_day = max(0.0, min(24.0, time_of_day))
        time_normalized = time_of_day / 24.0
        task_context_normalized["time_of_day"] = time_of_day

        self.current_task_context = task_context_normalized

        # Create status
        self.current_state = State(
            task_type=self._encode_task_type(task_type),
            tech_stack=tech_stack_vector,
            user_mood=user_mood_vector,
            time_of_day=time_normalized,
            recent_performance=self.recent_performance,
            agent_usage_history=self.agent_usage_history.copy()
        )

        return self.current_state

    def step(self, action: Action, task_result: Dict[str, Any]) -> Tuple[State, float, bool, Dict[str, Any]]:
        """
        perform action

        Args:
            action: action taken
            task_result: Task execution results，Include：
                - duration: Task time（Second）
                - test_result: Test results
                - user_feedback: User feedback
                - metrics: Other indicators

        Returns:
            (next_state, reward, done, info)
        """
        if self.current_state is None or self.current_task_context is None:
            raise ValueError("Environment not initialized, call reset() before step().")

        # 1. RecordAgentuse
        agent_name = action.agent_selection.value
        self.agent_usage_history[agent_name] += 1

        # 2. Prepare context for reward calculation
        reward_context = {
            "task_type": self.current_task_context.get("task_type", "T2"),
            "task_result": task_result,
            "test_result": task_result.get("test_result", {}),
            "user_feedback": task_result.get("user_feedback", {}),
            "metrics": task_result.get("metrics", {})
        }

        # 3. Calculate rewards
        reward = self.reward_calculator.calculate_reward(reward_context)

        # 4. Update recent performance（moving average）
        self.recent_performance = 0.7 * self.recent_performance + 0.3 * reward

        # 5. Prepare for next state（keep current context，But update performance）
        next_state = State(
            task_type=self.current_state.task_type.copy(),
            tech_stack=self.current_state.tech_stack.copy(),
            user_mood=self.current_state.user_mood.copy(),
            time_of_day=self.current_state.time_of_day,
            recent_performance=self.recent_performance,
            agent_usage_history=self.agent_usage_history.copy()
        )

        self.current_state = next_state

        # 6. examineepisodeIs it over?
        done = task_result.get("completed", True)
        if done:
            self.episode_count += 1
            self.episode_rewards.append(float(reward))

        # 7. Prepareinfo
        info = {
            "action_taken": str(action),
            "agent_used": agent_name,
            "reward_breakdown": reward_context.get("debug_info", {}).get("reward_breakdown", {}),
            "task_duration": task_result.get("duration", 0)
        }

        return next_state, reward, done, info

    def _encode_task_type(self, task_type: TaskType) -> np.ndarray:
        """Encode the task type asone-hotvector"""
        vector = np.zeros(self.STATE_DIM["task_type"])
        index = list(TaskType).index(task_type)
        vector[index] = 1.0
        return vector

    def _encode_tech_stack(self, tech_stacks: List[str]) -> np.ndarray:
        """Codify the technology stack asone-hotvector（Support multiple tags）"""
        vector = np.zeros(self.STATE_DIM["tech_stack"])

        for tech in tech_stacks:
            tech_lower = tech.lower()
            # Check for direct matches
            if tech_lower in self.SUPPORTED_TECH:
                vector[self.SUPPORTED_TECH[tech_lower]] = 1.0
            # Check for partial matches
            else:
                for supported_tech, idx in self.SUPPORTED_TECH.items():
                    if supported_tech in tech_lower:
                        vector[idx] = 1.0

        # If no technology matches，defaultpython
        if not vector.any():
            vector[self.SUPPORTED_TECH["python"]] = 1.0

        return vector

    def _encode_user_mood(self, user_mood: str) -> np.ndarray:
        """Encode user mood asone-hotvector"""
        moods = list(USER_MOOD_ORDER)
        vector = np.zeros(len(moods))

        if user_mood in moods:
            index = moods.index(user_mood)
            vector[index] = 1.0
        else:
            # defaultfocused
            vector[0] = 1.0

        return vector

    def get_action_space_size(self) -> int:
        """Get action space size"""
        return self.TOTAL_ACTION_DIM

    def get_state_space_size(self) -> int:
        """Get the state space size"""
        return self.TOTAL_STATE_DIM

    def save_history(self, path: str | Path) -> None:
        """Save history"""
        history = {
            "episode_count": self.episode_count,
            "agent_usage_history": self.agent_usage_history,
            "recent_performance": self.recent_performance,
            "reward_history": self.episode_rewards
        }

        history_path = Path(path).expanduser()
        history_path.parent.mkdir(parents=True, exist_ok=True)

        with open(history_path, 'w') as f:
            json.dump(history, f, indent=2)

    def _load_history(self, path: str | Path) -> None:
        """Load history"""
        history_path = Path(path).expanduser()

        if not history_path.exists():
            return

        with open(history_path, 'r') as f:
            history = json.load(f)

        self.episode_count = history.get("episode_count", 0)
        self.agent_usage_history = history.get("agent_usage_history", self.agent_usage_history)
        self.recent_performance = history.get("recent_performance", 0.5)
        self.episode_rewards = history.get("reward_history", [])


def main():
    """Test interactive environment"""
    # Create environment
    env = InteractionEnvironment()

    # Simulate task context
    task_context = {
        "task_type": "T2",
        "tech_stack": ["python", "fastapi"],
        "user_mood": "focused",
        "time_of_day": 14.0  # afternoon2point
    }

    # Reset environment
    state = env.reset(task_context)
    print(f"initial state: {state}")
    print(f"state vector: {state.to_vector()}")
    print(f"State space size: {env.get_state_space_size()}")
    print(f"action space size: {env.get_action_space_size()}")

    # simulate action
    action = Action(
        agent_selection=AgentType.CLAUDE,
        automation_level=AutomationLevel.MEDIUM,
        communication_style=CommunicationStyle.DETAILED,
        confirmation_needed=True
    )

    print(f"\ntake action: {action}")

    # Simulation task results
    task_result = {
        "duration": 300,  # 5minute
        "completed": True,
        "test_result": {
            "coverage": 85.0,
            "passed": 10,
            "failed": 1
        },
        "user_feedback": {
            "accepted": True,
            "rating": 4
        },
        "metrics": {
            "complexity": 3,
            "duplication": 0.05,
            "lint_score": 0.9
        }
    }

    # Execution steps
    next_state, reward, done, info = env.step(action, task_result)

    print(f"\nnext state: {next_state}")
    print(f"award: {reward:.3f}")
    print(f"Finish: {done}")
    print(f"Info: {info}")

    # save history
    env.save_history("/tmp/clawpolicy_env_history.json")
    print("\nHistory saved to /tmp/clawpolicy_env_history.json")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
reinforcement learning trainer - Complete training cycle

Implement a complete training process：
- manyepisodetrain
- checkpoint save/load
- Training statistics and visualization
"""

import numpy as np
from typing import Dict, List, Any, Optional, Tuple
from pathlib import Path
import json

from .agent import AlignmentAgent, Trajectory
from .environment import InteractionEnvironment
from .experience_replay import ExperienceReplay, Experience


class RLTrainer:
    """
    reinforcement learning trainer

    Function：
    - Complete training cycle（Multipleepisode）
    - checkpoint save/load
    - Training statistics and visualization
    """

    def __init__(self, model_dir: str | Path | None = None,
                 use_experience_replay: bool = True,
                 replay_capacity: int = 10000):
        """
        Initialize the trainer

        Args:
            model_dir: Model save directory
            use_experience_replay: Whether to use experience playback
            replay_capacity: Experience playback capacity
        """
        self.model_dir = Path(model_dir).expanduser() if model_dir else Path("./models/rl")
        self.model_dir.mkdir(parents=True, exist_ok=True)

        # Initialize environment
        self.env = InteractionEnvironment()

        # Initialize the agent
        self.agent = AlignmentAgent(
            state_dim=self.env.get_state_space_size(),
            action_dim=self.env.get_action_space_size()
        )

        # Initialize experience playback
        self.use_experience_replay = use_experience_replay
        self.replay_buffer = ExperienceReplay(capacity=replay_capacity) if use_experience_replay else None

        # training statistics
        self.episode_rewards: List[float] = []
        self.episode_lengths: List[int] = []
        self.training_losses: List[Dict[str, float]] = []

        # currentepisode
        self.current_episode = 0

    def train(self, num_episodes: int = 100,
              max_steps_per_episode: int = 100,
              save_interval: int = 10) -> Dict[str, Any]:
        """
        Train the agent

        Args:
            num_episodes: trainepisodequantity
            max_steps_per_episode: eachepisodeMaximum number of steps
            save_interval: save interval

        Returns:
            training statistics
        """
        print(f"🚀 Start training（{num_episodes} episodes）...")

        for episode in range(num_episodes):
            self.current_episode = episode + 1

            # run aepisode
            episode_reward, episode_length, episode_loss = self._run_episode(max_steps_per_episode)

            # Record statistics
            self.episode_rewards.append(episode_reward)
            self.episode_lengths.append(episode_length)
            self.training_losses.append(episode_loss)

            # Printing progress
            if episode % 10 == 0 or episode == num_episodes - 1:
                avg_reward = np.mean(self.episode_rewards[-10:])
                print(f"Episode {episode + 1}/{num_episodes} | "
                      f"award: {episode_reward:.3f} | "
                      f"average: {avg_reward:.3f} | "
                      f"number of steps: {episode_length}")

            # Save regularly
            if episode % save_interval == 0 and episode > 0:
                self.save_checkpoint(f"checkpoint_episode_{episode}")

        print("✅ Training completed！")

        # Save final model
        self.save_checkpoint("final")

        return self.get_training_stats()

    def _run_episode(self, max_steps: int) -> Tuple[float, int, Dict[str, float]]:
        """
        run aepisode

        Args:
            max_steps: Maximum number of steps

        Returns:
            (total reward, number of steps, Loss statistics)
        """
        # Random task context
        task_context = self._generate_random_task_context()

        # Reset environment
        state = self.env.reset(task_context)

        trajectory = Trajectory([], [], [], [], [])
        total_reward = 0.0

        for step in range(max_steps):
            # Select action
            action = self.agent.select_action(state, explore=True)

            # Simulation task results
            task_result = self._simulate_task_result()

            # Execution steps
            next_state, reward, done, info = self.env.step(action, task_result)

            action_indices = self.agent.encode_action_indices(action)

            # record track
            trajectory.states.append(state.to_vector())
            trajectory.actions.append(action_indices)
            trajectory.rewards.append(reward)
            trajectory.dones.append(done)
            trajectory.next_states.append(next_state.to_vector())

            # Add to experience replay
            if self.replay_buffer:
                exp = Experience(
                    state=state.to_vector(),
                    action=action_indices,
                    reward=reward,
                    next_state=next_state.to_vector(),
                    done=done,
                    priority=abs(reward)
                )
                self.replay_buffer.add(exp)

            total_reward += reward
            state = next_state

            if done:
                break

        # update strategy
        loss_stats = self.agent.update_policy(trajectory)

        # If you use experience replay and have enough experience，Make additional training updates
        if self.replay_buffer and self.replay_buffer.is_ready(min_size=32):
            additional_loss = self._train_from_replay()
            # Combined loss statistics
            if additional_loss:
                for key, value in additional_loss.items():
                    loss_stats[key] = loss_stats.get(key, 0) + value

        return total_reward, len(trajectory), loss_stats

    def _train_from_replay(self, num_updates: int = 4) -> Optional[Dict[str, float]]:
        """
        Train from experience replays

        Args:
            num_updates: Update times

        Returns:
            Loss statistics
        """
        if not self.replay_buffer:
            return None

        total_actor_loss = 0.0
        total_critic_loss = 0.0

        for _ in range(num_updates):
            states, actions, rewards, next_states, dones = self.replay_buffer.get_batch(batch_size=32)

            if len(states) == 0:
                continue

            # Using Experience Replay Data Update Strategies
            for i in range(len(states)):
                state = states[i]
                action_indices = actions[i]
                reward = rewards[i]
                next_state = next_states[i]
                done = dones[i]

                # Calculate target value
                if done:
                    target_value = reward
                else:
                    target_value = reward + self.agent.gamma * self.agent.value_net.forward(next_state)

                # Computational Advantage
                current_value = self.agent.value_net.forward(state)
                advantage = target_value - current_value

                # renewActorandCritic
                actor_loss = self.agent.policy_net.update(state, action_indices, advantage)
                critic_loss = self.agent.value_net.update(state, target_value)

                total_actor_loss += actor_loss
                total_critic_loss += critic_loss

        return {
            "actor_loss": total_actor_loss / (num_updates * 32),
            "critic_loss": total_critic_loss / (num_updates * 32)
        }

    def _generate_random_task_context(self) -> Dict[str, Any]:
        """Generate random task context"""
        task_types = ["T1", "T2", "T3", "T4"]
        tech_stacks = [["python"], ["javascript"], ["python", "fastapi"], ["react", "typescript"]]
        moods = ["focused", "relaxed", "stressed"]

        # usePythonofrandom.choiceinstead ofnumpy
        import random
        return {
            "task_type": random.choice(task_types),
            "tech_stack": random.choice(tech_stacks),
            "user_mood": random.choice(moods),
            "time_of_day": float(np.random.uniform(0, 24))
        }

    def _simulate_task_result(self) -> Dict[str, Any]:
        """Simulation task results"""
        # Randomly generate results
        coverage = np.random.uniform(50, 100)
        passed = int(np.random.uniform(5, 15))
        failed = int(np.random.uniform(0, 3))

        return {
            "duration": np.random.uniform(100, 600),
            "completed": True,
            "test_result": {
                "coverage": coverage,
                "passed": passed,
                "failed": failed
            },
            "user_feedback": {
                "accepted": np.random.random() > 0.2,
                "rating": np.random.randint(3, 6)
            },
            "metrics": {
                "complexity": np.random.uniform(1, 7),
                "duplication": np.random.uniform(0, 0.2),
                "lint_score": np.random.uniform(0.6, 1.0)
            }
        }

    def save_checkpoint(self, name: str) -> None:
        """
        Save checkpoint

        Args:
            name: Checkpoint name
        """
        checkpoint_dir = self.model_dir / name
        checkpoint_dir.mkdir(parents=True, exist_ok=True)

        # Save agent
        self.agent.save_model(str(checkpoint_dir))

        # save statistics
        stats = {
            "episode": self.current_episode,
            "episode_rewards": self.episode_rewards,
            "episode_lengths": self.episode_lengths,
            "training_losses": self.training_losses[-100:]  # recent100indivual
        }

        with open(checkpoint_dir / "training_stats.json", 'w') as f:
            json.dump(stats, f, indent=2)

        print(f"✅ Checkpoint saved: {checkpoint_dir}")

    def load_checkpoint(self, name: str) -> None:
        """
        Load checkpoint

        Args:
            name: Checkpoint name
        """
        checkpoint_dir = self.model_dir / name

        # Load the agent
        self.agent.load_model(str(checkpoint_dir))

        # Load statistics
        stats_path = checkpoint_dir / "training_stats.json"
        if stats_path.exists():
            with open(stats_path, 'r') as f:
                stats = json.load(f)

            self.current_episode = stats["episode"]
            self.episode_rewards = stats["episode_rewards"]
            self.episode_lengths = stats["episode_lengths"]
            self.training_losses = stats["training_losses"]

        print(f"✅ Checkpoint loaded: {checkpoint_dir}")

    def get_training_stats(self) -> Dict[str, Any]:
        """Get training statistics"""
        if not self.episode_rewards:
            return {}

        return {
            "total_episodes": len(self.episode_rewards),
            "average_reward": float(np.mean(self.episode_rewards)),
            "reward_std": float(np.std(self.episode_rewards)),
            "max_reward": float(np.max(self.episode_rewards)),
            "min_reward": float(np.min(self.episode_rewards)),
            "average_length": float(np.mean(self.episode_lengths)),
            "latest_reward": float(self.episode_rewards[-1]),
            "improvement": float(np.mean(self.episode_rewards[-10:]) - np.mean(self.episode_rewards[:10])) if len(self.episode_rewards) >= 20 else 0.0
        }


def main():
    """Test trainer"""
    trainer = RLTrainer(model_dir="/tmp/rl_trainer_test", use_experience_replay=True)

    print("✅ Trainer has been created")

    # train a fewepisode
    stats = trainer.train(num_episodes=5, max_steps_per_episode=10, save_interval=2)

    print("\n📊 training statistics:")
    for key, value in stats.items():
        print(f"  {key}: {value}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
reinforcement learning agent - Actor-Criticaccomplish

accomplishActor-Criticalgorithm，include：
- PolicyNetwork: policy network（Output action probability distribution）
- ValueNetwork: value network（Estimated status value）
- AlignmentAgent: Actor-Criticagent

Phase 1: pureNumPyaccomplish（linear model）
Phase 2: OptionalPyTorchaccomplish（neural network）
"""

import numpy as np
from typing import Dict, List, Tuple
from dataclasses import dataclass
import json
from pathlib import Path

from .contracts import (
    ACTION_HEAD_DIMS,
    ACTION_VECTOR_DIM,
    AGENT_ORDER,
    AUTOMATION_ORDER,
    CONFIRM_ORDER,
    STYLE_ORDER,
)
from .environment import State, Action, AgentType, AutomationLevel, CommunicationStyle


@dataclass
class Trajectory:
    """Trajectory data class"""
    states: List[np.ndarray]  # status sequence
    actions: List[np.ndarray]  # action sequence（index vector）
    rewards: List[float]  # reward sequence
    dones: List[bool]  # Done sign
    next_states: List[np.ndarray]  # next state sequence

    def __len__(self) -> int:
        return len(self.states)

    def __repr__(self) -> str:
        return f"Trajectory(length={len(self)}, total_reward={sum(self.rewards):.2f})"


class PolicyNetwork:
    """
    policy network - Output action probability distribution

    Phase 1: linear model（logits = state @ weights + bias）
    Phase 2: Optional neural network（PyTorch）
    """

    def __init__(self, state_dim: int, action_dim: int = ACTION_VECTOR_DIM, hidden_dim: int = 64):
        """
        Initialize policy network

        Args:
            state_dim: status dimension
            action_dim: action dimension
            hidden_dim: Hidden layer dimensions（Phase 2use）
        """
        self.state_dim = state_dim
        self.action_dim = action_dim
        self.hidden_dim = hidden_dim

        # Bull action space
        self.head_dims = ACTION_HEAD_DIMS.copy()

        # Phase 1: Linear model parameters（long）
        self.weights = {
            name: np.random.randn(state_dim, dim) * 0.01
            for name, dim in self.head_dims.items()
        }
        self.bias = {
            name: np.zeros(dim)
            for name, dim in self.head_dims.items()
        }

    def forward(self, state: np.ndarray) -> Dict[str, np.ndarray]:
        """forward propagation：Count each headlogits"""
        return {
            name: state @ self.weights[name] + self.bias[name]
            for name in self.head_dims
        }

    def softmax(self, logits: np.ndarray) -> np.ndarray:
        """Softmaxactivation function"""
        # numerical stability：Subtract the maximum value
        exp_logits = np.exp(logits - np.max(logits))
        return exp_logits / np.sum(exp_logits)

    def get_action_probs(self, state: np.ndarray) -> Dict[str, np.ndarray]:
        """Get action probability distribution（long）"""
        logits = self.forward(state)
        return {name: self.softmax(head_logits) for name, head_logits in logits.items()}

    def sample_action(self, state: np.ndarray, explore: bool = True) -> Tuple[np.ndarray, Dict[str, np.ndarray]]:
        """
        Sampling action

        Args:
            state: Current status
            explore: Whether to explore（epsilon-greedy）

        Returns:
            (action_indices, action_probs)
        """
        action_probs = self.get_action_probs(state)

        if not explore:
            # reasoning mode：greedy choice，Avoid random jitter in recommended actions
            action_indices = np.array([
                int(np.argmax(action_probs["agent"])),
                int(np.argmax(action_probs["automation"])),
                int(np.argmax(action_probs["style"])),
                int(np.argmax(action_probs["confirm"]))
            ], dtype=int)
        elif np.random.random() < 0.1:  # 10% epsilon-greedy
            # Random exploration（Each head is independently randomized）
            action_indices = np.array([
                np.random.randint(self.head_dims["agent"]),
                np.random.randint(self.head_dims["automation"]),
                np.random.randint(self.head_dims["style"]),
                np.random.randint(self.head_dims["confirm"])
            ], dtype=int)
        else:
            # Sampling by probability（Each head is sampled independently）
            action_indices = np.array([
                np.random.choice(self.head_dims["agent"], p=action_probs["agent"]),
                np.random.choice(self.head_dims["automation"], p=action_probs["automation"]),
                np.random.choice(self.head_dims["style"], p=action_probs["style"]),
                np.random.choice(self.head_dims["confirm"], p=action_probs["confirm"])
            ], dtype=int)

        return action_indices, action_probs

    def update(self, state: np.ndarray, action_indices: np.ndarray, advantage: float,
               learning_rate: float = 0.01) -> float:
        """
        Update policy network（REINFORCEalgorithm）

        Args:
            state: Current status
            action_indices: Action index executed（long）
            advantage: advantage function A(s,a) = Q(s,a) - V(s)
            learning_rate: learning rate

        Returns:
            loss value
        """
        action_probs = self.get_action_probs(state)

        total_loss = 0.0
        head_order = ["agent", "automation", "style", "confirm"]

        for head_idx, head_name in enumerate(head_order):
            probs = action_probs[head_name]
            action_idx = int(action_indices[head_idx])
            log_prob = np.log(probs[action_idx] + 1e-10)

            # Policy gradient: -log π(a|s) * A
            loss = -log_prob * advantage
            total_loss += loss

            one_hot = np.zeros_like(probs)
            one_hot[action_idx] = 1.0
            grad_logits = (one_hot - probs) * advantage

            # gradient ascent（Equivalent to loss reduction）
            self.weights[head_name] += learning_rate * np.outer(state, grad_logits)
            self.bias[head_name] += learning_rate * grad_logits

        return float(total_loss)

    def save(self, path: str | Path) -> None:
        """Save model parameters"""
        params = {
            "weights": {name: w.tolist() for name, w in self.weights.items()},
            "bias": {name: b.tolist() for name, b in self.bias.items()},
            "state_dim": self.state_dim,
            "action_dim": self.action_dim,
            "head_dims": self.head_dims
        }

        model_path = Path(path).expanduser()
        model_path.parent.mkdir(parents=True, exist_ok=True)

        with open(model_path, 'w') as f:
            json.dump(params, f)

    def load(self, path: str | Path) -> None:
        """Load model parameters"""
        model_path = Path(path).expanduser()

        if not model_path.exists():
            return

        with open(model_path, 'r') as f:
            params = json.load(f)

        self.weights = {name: np.array(w) for name, w in params["weights"].items()}
        self.bias = {name: np.array(b) for name, b in params["bias"].items()}
        self.state_dim = params["state_dim"]
        self.action_dim = params["action_dim"]
        self.head_dims = params.get("head_dims", self.head_dims)


class ValueNetwork:
    """
    value network - Estimated status value V(s)

    Phase 1: linear model
    Phase 2: Optional neural network（PyTorch）
    """

    def __init__(self, state_dim: int, hidden_dim: int = 64):
        """
        Initialize the value network

        Args:
            state_dim: status dimension
            hidden_dim: Hidden layer dimensions（Phase 2use）
        """
        self.state_dim = state_dim
        self.hidden_dim = hidden_dim

        # Linear model parameters
        self.weights = np.random.randn(state_dim) * 0.01
        self.bias = 0.0

    def forward(self, state: np.ndarray) -> float:
        """forward propagation：Calculate status value"""
        return float(state @ self.weights + self.bias)

    def update(self, state: np.ndarray, target_value: float,
               learning_rate: float = 0.01) -> float:
        """
        Update value network（MSEloss）

        Args:
            state: Current status
            target_value: target value（actual return）
            learning_rate: learning rate

        Returns:
            loss value
        """
        # Calculate current value
        current_value = self.forward(state)

        # Calculate losses
        loss = (current_value - target_value) ** 2

        # Calculate gradient
        grad_w = 2 * (current_value - target_value) * state
        grad_b = 2 * (current_value - target_value)

        # Update weights
        self.weights -= learning_rate * grad_w
        self.bias -= learning_rate * grad_b

        return loss

    def save(self, path: str | Path) -> None:
        """Save model parameters"""
        params = {
            "weights": self.weights.tolist(),
            "bias": float(self.bias),
            "state_dim": self.state_dim
        }

        model_path = Path(path).expanduser()
        model_path.parent.mkdir(parents=True, exist_ok=True)

        with open(model_path, 'w') as f:
            json.dump(params, f)

    def load(self, path: str | Path) -> None:
        """Load model parameters"""
        model_path = Path(path).expanduser()

        if not model_path.exists():
            return

        with open(model_path, 'r') as f:
            params = json.load(f)

        self.weights = np.array(params["weights"])
        self.bias = params["bias"]
        self.state_dim = params["state_dim"]


class AlignmentAgent:
    """
    Actor-Criticagent

    Combining strategy network and value network，accomplishActor-Criticalgorithm：
    - Actor: policy network，Select action
    - Critic: value network，Evaluate state value

    Algorithm process：
    1. Select an action using the current strategy
    2. perform action，Get rewards and next status
    3. Compute advantage function A = R + γV(s') - V(s)
    4. renewActor：-log π(a|s) * A
    5. renewCritic：(V(s) - R)²
    """

    def __init__(self, state_dim: int, action_dim: int,
                 gamma: float = 0.99,
                 actor_lr: float = 0.01,
                 critic_lr: float = 0.01):
        """
        Initialize the agent

        Args:
            state_dim: status dimension
            action_dim: action dimension
            gamma: discount factor
            actor_lr: Actorlearning rate
            critic_lr: Criticlearning rate
        """
        self.state_dim = state_dim
        self.action_dim = action_dim
        self.gamma = gamma

        # Initialize the network
        self.policy_net = PolicyNetwork(state_dim, action_dim)
        self.value_net = ValueNetwork(state_dim)

        # learning rate
        self.actor_lr = actor_lr
        self.critic_lr = critic_lr

        # training statistics
        self.episode_count = 0
        self.total_steps = 0

    def select_action(self, state: State, explore: bool = True) -> Action:
        """
        Select action

        Args:
            state: Current status
            explore: Whether to explore

        Returns:
            selected action
        """
        # Convert state to vector
        state_vector = state.to_vector()

        # Sampling action index
        action_indices, action_probs = self.policy_net.sample_action(state_vector, explore)

        # Convert action index toActionobject
        action = self.decode_action_indices(action_indices)

        return action

    def encode_action_indices(self, action: Action) -> np.ndarray:
        """WillActionencoded as index vector"""
        agent_idx = AGENT_ORDER.index(action.agent_selection.value)
        automation_idx = AUTOMATION_ORDER.index(action.automation_level.value)
        style_idx = STYLE_ORDER.index(action.communication_style.value)
        confirm_idx = CONFIRM_ORDER.index(action.confirmation_needed)

        return np.array([agent_idx, automation_idx, style_idx, confirm_idx], dtype=int)

    def decode_action_indices(self, action_indices: np.ndarray) -> Action:
        """Decode action index asActionobject"""
        indices = [int(x) for x in action_indices]
        if len(indices) != 4:
            raise ValueError(f"action_indices must contain 4 values, got {len(indices)}")
        agent_idx, automation_idx, style_idx, confirm_idx = indices

        if not (0 <= agent_idx < len(AGENT_ORDER)):
            raise ValueError(f"agent index out of range: {agent_idx}")
        if not (0 <= automation_idx < len(AUTOMATION_ORDER)):
            raise ValueError(f"automation index out of range: {automation_idx}")
        if not (0 <= style_idx < len(STYLE_ORDER)):
            raise ValueError(f"style index out of range: {style_idx}")
        if not (0 <= confirm_idx < len(CONFIRM_ORDER)):
            raise ValueError(f"confirm index out of range: {confirm_idx}")

        agent_type = AgentType(AGENT_ORDER[agent_idx])
        automation = AutomationLevel(AUTOMATION_ORDER[automation_idx])
        style = CommunicationStyle(STYLE_ORDER[style_idx])
        confirm = CONFIRM_ORDER[confirm_idx]

        return Action(
            agent_selection=agent_type,
            automation_level=automation,
            communication_style=style,
            confirmation_needed=confirm
        )

    def update_policy(self, trajectory: Trajectory) -> Dict[str, float]:
        """
        update strategy（Actor-Criticalgorithm）

        Args:
            trajectory: complete trajectory

        Returns:
            Loss statistics
        """
        if len(trajectory) == 0:
            return {}

        total_actor_loss = 0.0
        total_critic_loss = 0.0

        # Gradually update
        for i in range(len(trajectory)):
            state = trajectory.states[i]
            action_indices = trajectory.actions[i]
            reward = trajectory.rewards[i]
            next_state = trajectory.next_states[i]
            done = trajectory.dones[i]

            # Calculate target value
            if done:
                target_value = reward
            else:
                target_value = reward + self.gamma * self.value_net.forward(next_state)

            # Compute advantage function
            current_value = self.value_net.forward(state)
            advantage = target_value - current_value

            # renewActor
            actor_loss = self.policy_net.update(state, action_indices, advantage, self.actor_lr)
            total_actor_loss += actor_loss

            # renewCritic
            critic_loss = self.value_net.update(state, target_value, self.critic_lr)
            total_critic_loss += critic_loss

        # Update statistics
        self.episode_count += 1
        self.total_steps += len(trajectory)

        return {
            "actor_loss": total_actor_loss / len(trajectory),
            "critic_loss": total_critic_loss / len(trajectory),
            "episode_length": len(trajectory),
            "total_return": sum(trajectory.rewards)
        }

    def _compute_returns(self, rewards: List[float], dones: List[bool]) -> List[float]:
        """
        Calculate discounted returns

        Args:
            rewards: reward sequence
            dones: Complete flag sequence

        Returns:
            discount return sequence
        """
        returns: List[float] = []
        running_return = 0.0

        # Calculate from back to front
        for reward, done in zip(reversed(rewards), reversed(dones)):
            if done:
                running_return = reward
            else:
                running_return = reward + self.gamma * running_return

            returns.insert(0, running_return)

        return returns

    def save_model(self, path: str) -> None:
        """Save model"""
        model_dir = Path(path).expanduser()
        model_dir.mkdir(parents=True, exist_ok=True)

        # save policy network
        self.policy_net.save(str(model_dir / "policy_network.json"))

        # Save value network
        self.value_net.save(str(model_dir / "value_network.json"))

        # Save metadata
        metadata = {
            "episode_count": self.episode_count,
            "total_steps": self.total_steps,
            "gamma": self.gamma,
            "actor_lr": self.actor_lr,
            "critic_lr": self.critic_lr
        }

        with open(model_dir / "metadata.json", 'w') as f:
            json.dump(metadata, f, indent=2)

    def load_model(self, path: str) -> None:
        """Load model"""
        model_dir = Path(path).expanduser()

        # Load policy network
        self.policy_net.load(str(model_dir / "policy_network.json"))

        # Load value network
        self.value_net.load(str(model_dir / "value_network.json"))

        # Load metadata
        metadata_path = model_dir / "metadata.json"
        if metadata_path.exists():
            with open(metadata_path, 'r') as f:
                metadata = json.load(f)

            self.episode_count = metadata.get("episode_count", 0)
            self.total_steps = metadata.get("total_steps", 0)


def main():
    """test agent"""
    from .environment import InteractionEnvironment

    # Create environments and agents
    env = InteractionEnvironment()
    agent = AlignmentAgent(
        state_dim=env.get_state_space_size(),
        action_dim=env.get_action_space_size()
    )

    print("Agent has been created")
    print(f"state space: {env.get_state_space_size()}")
    print(f"action space: {env.get_action_space_size()}")

    # Simulation training
    for episode in range(3):
        print(f"\n=== Episode {episode + 1} ===")

        # Reset environment
        task_context = {
            "task_type": "T2",
            "tech_stack": ["python"],
            "user_mood": "focused"
        }

        state = env.reset(task_context)

        # Collect tracks
        trajectory = Trajectory([], [], [], [], [])

        for step in range(5):  # eachepisodemost5step
            # Select action
            action = agent.select_action(state, explore=True)

            # Simulation task results
            task_result = {
                "duration": 200 + step * 50,
                "completed": step == 4,  # Last step completed
                "test_result": {"coverage": 70 + step * 5},
                "user_feedback": {"accepted": True},
                "metrics": {}
            }

            # Execution steps
            next_state, reward, done, info = env.step(action, task_result)

            # record track
            trajectory.states.append(state.to_vector())
            trajectory.actions.append(agent.encode_action_indices(action))
            trajectory.rewards.append(reward)
            trajectory.dones.append(done)
            trajectory.next_states.append(next_state.to_vector())

            state = next_state

            print(f"  Step {step + 1}: reward={reward:.3f}, done={done}")

            if done:
                break

        # update strategy
        stats = agent.update_policy(trajectory)
        print(f"  total return: {stats['total_return']:.3f}")
        print(f"  Actorloss: {stats['actor_loss']:.4f}")
        print(f"  Criticloss: {stats['critic_loss']:.4f}")

    # Save model
    agent.save_model("/tmp/clawpolicy_rl_agent")
    print("\nModel saved to /tmp/clawpolicy_rl_agent")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Tests for the skills health check script.
"""

from __future__ import annotations

import shutil
from collections.abc import Sequence
from pathlib import Path

import scripts.check_skills_health as skills_health


def _write_skill(path: Path) -> None:
    path.mkdir(parents=True, exist_ok=True)
    (path / "SKILL.md").write_text("# skill\n", encoding="utf-8")


def test_scan_detects_broken_symlink_and_missing_skill_md(tmp_path: Path) -> None:
    skills_root = tmp_path / ".codex" / "skills"
    skills_root.mkdir(parents=True)
    _write_skill(skills_root / "healthy-skill")

    broken_target = tmp_path / ".agents" / "skills" / "missing-source" / "skills" / "ghost-skill"
    (skills_root / "ghost-skill").symlink_to(broken_target)

    (skills_root / "invalid-skill").mkdir()

    report = skills_health.run_health_check(
        skill_roots=[skills_root],
        agents_root=tmp_path / ".agents" / "skills",
        source_repo_map={},
        relink_rules={},
        repair=False,
    )

    issue_codes = {issue.code for issue in report.issues}
    assert "BROKEN_SYMLINK" in issue_codes
    assert "MISSING_SKILL_MD" in issue_codes


def test_scan_ignores_support_directories_without_skill_md(tmp_path: Path) -> None:
    skills_root = tmp_path / ".codex" / "skills"
    skills_root.mkdir(parents=True)
    (skills_root / "assets").mkdir()
    (skills_root / "templates").mkdir()

    report = skills_health.run_health_check(
        skill_roots=[skills_root],
        agents_root=tmp_path / ".agents" / "skills",
        source_repo_map={},
        relink_rules={},
        repair=False,
    )

    assert not report.issues


def test_repair_clones_missing_source_repo_and_recovers_symlink(tmp_path: Path) -> None:
    skills_root = tmp_path / ".codex" / "skills"
    skills_root.mkdir(parents=True)

    agents_root = tmp_path / ".agents" / "skills"
    broken_target = agents_root / "skill-from-masters" / "skills" / "search-skill"
    broken_link = skills_root / "search-skill"
    broken_link.symlink_to(broken_target)

    source_template = tmp_path / "source-template"
    _write_skill(source_template / "skills" / "search-skill")

    def fake_runner(command: Sequence[str]) -> int:
        assert command[:4] == ["git", "clone", "--depth", "1"]
        destination = Path(command[5])
        shutil.copytree(source_template, destination)
        return 0

    report = skills_health.run_health_check(
        skill_roots=[skills_root],
        agents_root=agents_root,
        source_repo_map={"skill-from-masters": "https://example.invalid/skill-from-masters.git"},
        relink_rules={},
        repair=True,
        command_runner=fake_runner,
    )

    assert not report.issues
    assert report.cloned_repos == ["skill-from-masters"]
    assert broken_link.exists()
    assert (broken_link / "SKILL.md").is_file()


def test_repair_relinks_find_skills_to_known_target(tmp_path: Path) -> None:
    claude_skills = tmp_path / ".claude" / "skills"
    claude_skills.mkdir(parents=True)
    agents_root = tmp_path / ".agents" / "skills"

    broken_link = claude_skills / "find-skills"
    broken_link.symlink_to(agents_root / "find-skills")

    fixed_target = agents_root / "vercel-labs-skills" / "skills" / "find-skills"
    _write_skill(fixed_target)

    report = skills_health.run_health_check(
        skill_roots=[claude_skills],
        agents_root=agents_root,
        source_repo_map={},
        relink_rules={"find-skills": fixed_target},
        repair=True,
    )

    assert not report.issues
    assert broken_link.resolve() == fixed_target.resolve()
    assert any("find-skills" in action for action in report.repaired_actions)



#!/usr/bin/env python3
"""
RLLearnerhistorical preference test
"""

import pytest

from lib.learner import RLLearner


def test_rl_learner_records_preference_history(tmp_path):
    learner = RLLearner(
        model_path=str(tmp_path / "models"),
        config_path=str(tmp_path / "config.json")
    )

    assert learner.get_agent_success_rate("T2", "claude") is None
    assert learner.get_workflow_success_rate("T2", "tdd") is None

    learner.record_preference_result(task_type="T2", agent="claude", workflow="tdd", reward=0.9)
    learner.record_preference_result(task_type="T2", agent="claude", workflow="tdd", reward=0.3)

    assert learner.get_agent_success_rate("T2", "claude") == pytest.approx(0.6)
    assert learner.get_workflow_success_rate("T2", "tdd") == pytest.approx(0.6)
    assert learner.env.reward_calculator.history_provider is learner



#!/usr/bin/env python3
"""Release readiness regression tests."""

from __future__ import annotations

import re
import subprocess
import sys
from pathlib import Path


def _repo_root() -> Path:
    return Path(__file__).resolve().parent.parent


def test_pyproject_readme_points_to_existing_file() -> None:
    repo_root = _repo_root()
    content = (repo_root / "pyproject.toml").read_text(encoding="utf-8")
    match = re.search(r'^readme\s*=\s*"([^"]+)"\s*$', content, flags=re.MULTILINE)
    assert match is not None, "pyproject.toml is missing the readme field."

    readme_path = repo_root / match.group(1)
    assert readme_path.exists(), f"readme file does not exist: {readme_path.name}"


def test_pyproject_exposes_public_cli_entrypoint() -> None:
    repo_root = _repo_root()
    content = (repo_root / "pyproject.toml").read_text(encoding="utf-8")
    assert 'name = "clawpolicy"' in content
    assert 'clawpolicy = "clawpolicy.cli:main"' in content
    scripts_section = re.search(r"^\[project\.scripts\]$(.*?)(^\[|\Z)", content, flags=re.MULTILINE | re.DOTALL)
    assert scripts_section is not None, "pyproject.toml is missing the project.scripts section."
    script_lines = [
        line.strip()
        for line in scripts_section.group(1).splitlines()
        if line.strip() and not line.strip().startswith("#")
    ]
    assert script_lines == ['clawpolicy = "clawpolicy.cli:main"']


def test_import_lib_works_without_torch_installed() -> None:
    repo_root = _repo_root()
    script = (
        "import importlib, sys\n"
        "sys.modules['torch'] = None\n"
        "sys.modules.pop('torch.nn', None)\n"
        "sys.modules.pop('torch.nn.functional', None)\n"
        "importlib.import_module('lib')\n"
        "print('ok')\n"
    )
    result = subprocess.run(
        [sys.executable, "-c", script],
        cwd=repo_root,
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode == 0, result.stdout + result.stderr
    assert "ok" in result.stdout


def test_ci_workflow_covers_master_and_phase3_has_pytest() -> None:
    repo_root = _repo_root()
    content = (repo_root / ".github/workflows/ci.yml").read_text(encoding="utf-8")

    assert "master" in content, "CI push branches do not include master."

    assert "phase3-smoke" in content
    assert "phase3-full-deps" in content
    phase3_install_has_pytest = (
        "python -m pip install -r requirements-dev.txt" in content
        or "python -m pip install pytest" in content
    )
    assert phase3_install_has_pytest, "phase3 job does not install pytest."


def test_release_workflow_has_pypi_token_fallback() -> None:
    repo_root = _repo_root()
    content = (repo_root / ".github/workflows/release.yml").read_text(encoding="utf-8")
    assert "PYPI_API_TOKEN" in content, "release workflow is missing the PYPI_API_TOKEN fallback path."



#!/usr/bin/env python3
"""Integration tests for policy memory, CLI, and markdown conversion."""

from __future__ import annotations

import json
import tempfile
from datetime import datetime, timezone
from pathlib import Path
from shutil import rmtree

import pytest

from lib.cli import ClawPolicyCLI
from lib.md_to_policy import MarkdownToPolicyConverter
from lib.policy_store import PolicyStore
from lib.policy_to_md import PolicyToMarkdownExporter


class TestCLIInit:
    """Tests for `clawpolicy init`."""

    @pytest.fixture
    def temp_dir(self) -> Path:
        temp_dir = Path(tempfile.mkdtemp())
        yield temp_dir
        if temp_dir.exists():
            rmtree(temp_dir)

    def test_init_creates_policy_files(self, temp_dir: Path) -> None:
        cli = ClawPolicyCLI()
        cli.memory_dir_name = ".clawpolicy_test"

        import os

        original_cwd = os.getcwd()
        try:
            os.chdir(temp_dir)
            success = cli.init(force=True)
            assert success is True

            memory_dir = cli.get_memory_dir()
            policy_dir = memory_dir / "policy"
            assert policy_dir.exists()
            assert (policy_dir / "rules.json").exists()
            assert (policy_dir / "playbooks.json").exists()
            assert (policy_dir / "policy_events.jsonl").exists()
        finally:
            os.chdir(original_cwd)

    def test_init_writes_timestamp_and_working_directory_fields(self, temp_dir: Path) -> None:
        cli = ClawPolicyCLI()
        cli.memory_dir_name = ".clawpolicy_test"

        import os

        original_cwd = os.getcwd()
        try:
            os.chdir(temp_dir)
            success = cli.init(force=True)
            assert success is True

            config_path = cli.get_memory_dir() / cli.config_file_name
            config = json.loads(config_path.read_text(encoding="utf-8"))

            initialized_at = datetime.fromisoformat(config["initialized_at"])
            assert initialized_at.tzinfo is not None
            assert initialized_at.utcoffset() == timezone.utc.utcoffset(initialized_at)
            assert Path(config["initialized_in"]).resolve() == temp_dir.resolve()
            assert config["memory_path"] == str(cli.get_memory_dir())
        finally:
            os.chdir(original_cwd)

    def test_init_auto_migrates_markdown_into_policy_assets(self, temp_dir: Path) -> None:
        cli = ClawPolicyCLI()
        cli.memory_dir_name = ".clawpolicy_test"

        import os

        original_cwd = os.getcwd()
        try:
            os.chdir(temp_dir)
            memory_dir = cli.get_memory_dir()
            if memory_dir.exists():
                rmtree(memory_dir)
            memory_dir.mkdir(parents=True, exist_ok=True)

            (memory_dir / "USER.md").write_text(
                """# USER

## Basic Information

- Name: Test User
- Role: Developer

## Working Preferences

- Communication style: concise
- Automation preference: high
""",
                encoding="utf-8",
            )

            (memory_dir / "SOUL.md").write_text(
                """# SOUL

## Core Principles

1. Safety First

- Protect user data
""",
                encoding="utf-8",
            )

            success = cli.init(force=False)
            assert success is True

            policy_dir = memory_dir / "policy"
            store = PolicyStore(policy_dir)

            assert policy_dir.exists()
            assert store.load_rules() or store.load_playbooks() or store.get_events()
        finally:
            os.chdir(original_cwd)


class TestMarkdownToPolicyConversion:
    """Tests for markdown -> policy conversion."""

    @pytest.fixture
    def temp_dir(self) -> Path:
        temp_dir = Path(tempfile.mkdtemp())
        yield temp_dir
        if temp_dir.exists():
            rmtree(temp_dir)

    def test_convert_user_md(self, temp_dir: Path) -> None:
        user_md = temp_dir / "USER.md"
        user_md.write_text(
            """# USER

## Basic Information

- Name: Test User
- Role: Developer

## Working Preferences

- Communication style: concise
- Automation preference: high
""",
            encoding="utf-8",
        )

        converter = MarkdownToPolicyConverter()
        rules = converter.convert_user_md_to_rules(user_md)

        assert rules
        assert any("basic_info" in rule_id for rule_id in rules)

    def test_convert_soul_md(self, temp_dir: Path) -> None:
        soul_md = temp_dir / "SOUL.md"
        soul_md.write_text(
            """# SOUL

## Core Principles

1. Safety First

- Protect user data

## Prohibited Actions

- Destructive operations
""",
            encoding="utf-8",
        )

        converter = MarkdownToPolicyConverter()
        playbook = converter.convert_soul_md_to_playbook(soul_md)

        assert playbook is not None
        assert playbook.category == "harden"

    def test_convert_agents_md(self, temp_dir: Path) -> None:
        agents_md = temp_dir / "AGENTS.md"
        agents_md.write_text(
            """# AGENTS

## Tool Dispatch

- Codex: backend logic
- Claude: UI tasks

## Operation Rules

- Behavior changes require tests
""",
            encoding="utf-8",
        )

        converter = MarkdownToPolicyConverter()
        rules = converter.convert_agents_md_to_rules(agents_md)

        assert rules


class TestPolicyToMarkdownExport:
    """Tests for policy memory -> markdown export."""

    @pytest.fixture
    def temp_dir(self) -> Path:
        temp_dir = Path(tempfile.mkdtemp())
        yield temp_dir
        if temp_dir.exists():
            rmtree(temp_dir)

    def test_export_rules_to_user_md(self, temp_dir: Path) -> None:
        from lib.policy_models import Rule

        basic_rule = Rule(
            id="rule_basic_info",
            summary="Basic information",
            category="optimize",
            strategy="- Name: Test User\n- Role: Developer",
        )

        rules = {"rule_basic_info": basic_rule}

        exporter = PolicyToMarkdownExporter()
        output_path = temp_dir / "USER.md"
        exporter.export_rules_to_user_md(rules, output_path)

        assert output_path.exists()
        content = output_path.read_text(encoding="utf-8")
        assert "# USER" in content
        assert "Test User" in content

    def test_export_playbook_to_soul_md(self, temp_dir: Path) -> None:
        from lib.policy_models import Playbook

        playbook = Playbook(
            id="playbook_safety",
            summary="Safety boundary",
            category="harden",
            rules_used=[],
        )

        exporter = PolicyToMarkdownExporter()
        output_path = temp_dir / "SOUL.md"
        exporter.export_playbook_to_soul_md(playbook, output_path)

        assert output_path.exists()
        content = output_path.read_text(encoding="utf-8")
        assert "# SOUL" in content



#!/usr/bin/env python3
"""Cross-platform application paths for ClawPolicy."""

from __future__ import annotations

import os
import sys
from pathlib import Path
from typing import Optional, Union


APP_NAME = "clawpolicy"


def _home() -> Path:
    return Path.home()


def get_config_dir() -> Path:
    if sys.platform.startswith("win"):
        base = Path(os.environ.get("APPDATA", _home() / "AppData/Roaming"))
        return base / APP_NAME
    base = Path(os.environ.get("XDG_CONFIG_HOME", _home() / ".config"))
    return base / APP_NAME


def get_cache_dir() -> Path:
    if sys.platform.startswith("win"):
        base = Path(os.environ.get("LOCALAPPDATA", _home() / "AppData/Local"))
        return base / APP_NAME / "cache"
    base = Path(os.environ.get("XDG_CACHE_HOME", _home() / ".cache"))
    return base / APP_NAME


def get_state_dir() -> Path:
    if sys.platform.startswith("win"):
        base = Path(os.environ.get("LOCALAPPDATA", _home() / "AppData/Local"))
        return base / APP_NAME / "state"
    base = Path(os.environ.get("XDG_STATE_HOME", _home() / ".local/state"))
    return base / APP_NAME


def get_default_config_path() -> Path:
    return get_config_dir() / "config.json"


def get_default_model_dir() -> Path:
    return get_cache_dir() / "models" / "rl"


def resolve_config_path(config_path: Optional[Union[str, Path]] = None) -> Path:
    if config_path:
        return Path(config_path).expanduser()
    return get_default_config_path()


def resolve_model_dir(model_path: Optional[Union[str, Path]] = None) -> Path:
    if model_path:
        return Path(model_path).expanduser()
    return get_default_model_dir()



"""
ClawPolicy - explainable confirmation and local policy memory.

Phase 1-2: confirmation policy core, local policy memory, and optional RL optimization
Phase 3: advanced optional modules (distributed training, tuning, monitoring, performance)
"""

from __future__ import annotations

from typing import Any

# Phase 1: core module（always available）
from .reward import RewardSignal, RewardCalculator
from .environment import InteractionEnvironment, State, Action
from .agent import AlignmentAgent, PolicyNetwork, ValueNetwork
from .learner import RLLearner
from .experience_replay import Experience, ExperienceReplay
from .trainer import RLTrainer
from .paths import (
    get_config_dir,
    get_cache_dir,
    get_state_dir,
    resolve_config_path,
    resolve_model_dir,
)

# Policy memory module
from .policy_models import Rule, Playbook, PolicyEvent
from .policy_store import PolicyStore
from .md_to_policy import MarkdownToPolicyConverter
from .policy_to_md import PolicyToMarkdownExporter

# Intelligent Confirmation module
from .confirmation import IntelligentConfirmation, RiskLevel

# API module
from .api import ConfirmationAPI, create_api

# CLI module
from .cli import ClawPolicyCLI

# Optional exports（Injection by dependency availability at runtime）
MLPModel: Any = None
PolicyNetworkPyTorch: Any = None
ValueNetworkPyTorch: Any = None
create_policy_network: Any = None
create_value_network: Any = None
TORCH_AVAILABLE = False

DistributedTrainer: Any = None
DistributedTrainingConfig: Any = None

LearningRateScheduler: Any = None
HyperparameterSearch: Any = None
EarlyStopping: Any = None
HyperparameterTuner: Any = None

TrainingMonitor: Any = None
MetricsAnalyzer: Any = None
TENSORBOARD_AVAILABLE = False

BatchInference: Any = None
ModelQuantization: Any = None
InferenceCache: Any = None
JITOptimizer: Any = None
PerformanceOptimizer: Any = None

try:
    from . import nn_model as _nn_model

    MLPModel = _nn_model.MLPModel
    PolicyNetworkPyTorch = _nn_model.PolicyNetworkPyTorch
    ValueNetworkPyTorch = _nn_model.ValueNetworkPyTorch
    create_policy_network = _nn_model.create_policy_network
    create_value_network = _nn_model.create_value_network
    TORCH_AVAILABLE = _nn_model.TORCH_AVAILABLE
except Exception:
    pass

try:
    from . import distributed_trainer as _distributed_trainer

    DistributedTrainer = _distributed_trainer.DistributedTrainer
    DistributedTrainingConfig = _distributed_trainer.DistributedTrainingConfig
except Exception:
    pass

try:
    from . import hyperparameter_tuner as _hyperparameter_tuner

    LearningRateScheduler = _hyperparameter_tuner.LearningRateScheduler
    HyperparameterSearch = _hyperparameter_tuner.HyperparameterSearch
    EarlyStopping = _hyperparameter_tuner.EarlyStopping
    HyperparameterTuner = _hyperparameter_tuner.HyperparameterTuner
except Exception:
    pass

try:
    from . import monitoring as _monitoring

    TrainingMonitor = _monitoring.TrainingMonitor
    MetricsAnalyzer = _monitoring.MetricsAnalyzer
    TENSORBOARD_AVAILABLE = _monitoring.TENSORBOARD_AVAILABLE
except Exception:
    pass

try:
    from . import performance_optimizer as _performance_optimizer

    BatchInference = _performance_optimizer.BatchInference
    ModelQuantization = _performance_optimizer.ModelQuantization
    InferenceCache = _performance_optimizer.InferenceCache
    JITOptimizer = _performance_optimizer.JITOptimizer
    PerformanceOptimizer = _performance_optimizer.PerformanceOptimizer
except Exception:
    pass

__all__ = [
    # Phase 1
    "RewardSignal",
    "RewardCalculator",
    "InteractionEnvironment",
    "State",
    "Action",
    "AlignmentAgent",
    "PolicyNetwork",
    "ValueNetwork",
    "RLLearner",
    "RLAlignmentEngine",
    "Experience",
    "ExperienceReplay",
    "RLTrainer",
    "get_config_dir",
    "get_cache_dir",
    "get_state_dir",
    "resolve_config_path",
    "resolve_model_dir",
    # Policy memory
    "Rule",
    "Playbook",
    "PolicyEvent",
    "PolicyStore",
    "MarkdownToPolicyConverter",
    "PolicyToMarkdownExporter",
    # Intelligent Confirmation
    "IntelligentConfirmation",
    "RiskLevel",
    # API
    "ConfirmationAPI",
    "create_api",
    # Phase 2
    "MLPModel",
    "PolicyNetworkPyTorch",
    "ValueNetworkPyTorch",
    "create_policy_network",
    "create_value_network",
    "TORCH_AVAILABLE",
    # Phase 3
    "DistributedTrainer",
    "DistributedTrainingConfig",
    "LearningRateScheduler",
    "HyperparameterSearch",
    "EarlyStopping",
    "HyperparameterTuner",
    "TrainingMonitor",
    "MetricsAnalyzer",
    "TENSORBOARD_AVAILABLE",
    "BatchInference",
    "ModelQuantization",
    "InferenceCache",
    "JITOptimizer",
    "PerformanceOptimizer",
    # CLI
    "ClawPolicyCLI",
]


def __getattr__(name: str) -> Any:
    if name == "RLAlignmentEngine":
        # Delayed import，avoid `python -m lib.integration` Warning when module is loaded repeatedly
        from .integration import RLAlignmentEngine as _RLAlignmentEngine

        return _RLAlignmentEngine
    raise AttributeError(f"module 'lib' has no attribute '{name}'")


__version__ = "3.0.0"
